from typing import Callable, Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from dash import Dash, Input, Output, State, no_update
from dash.exceptions import PreventUpdate
from webviz_config.common_cache import CACHE

from webviz_subsurface._models import ObservationModel
from webviz_subsurface._providers import EnsembleTableProvider

from ..figures.plotly_line_plot import PlotlyLinePlot


def build_figure(
    app: Dash,
    get_uuid: Callable,
    tableproviders: Dict[str, EnsembleTableProvider],
    observationmodel: Optional[ObservationModel],
    parameterproviders: Dict[str, EnsembleTableProvider],
    colors: Dict,
) -> None:
    @app.callback(
        Output(
            {"id": get_uuid("clientside"), "plotly_attribute": "plotly_data"}, "data"
        ),
        Output(
            {"id": get_uuid("clientside"), "plotly_attribute": "initial_layout"}, "data"
        ),
        Input(get_uuid("stored_x_value"), "data"),
        Input(
            {
                "id": get_uuid("data_selectors"),
                "data_attribute": "y",
                "source": "table",
            },
            "value",
        ),
        Input(
            {
                "id": get_uuid("data_selectors"),
                "data_attribute": "ensemble",
                "source": "table",
            },
            "value",
        ),
        Input(
            {
                "id": get_uuid("data_selectors"),
                "data_attribute": "parameter",
                "source": "parameter",
            },
            "value",
        ),
        Input(get_uuid("traces"), "value"),
        Input(get_uuid("observations"), "value"),
        Input(get_uuid("highlight-realizations"), "value"),
        Input(get_uuid("mode"), "value"),
        Input({"id": get_uuid("parameter-filter"), "type": "data-store"}, "data"),
    )
    # pylint: disable=too-many-locals, too-many-arguments
    def _update_plot(
        x_column_name: str,
        y_column_name: str,
        ensemble_names: Union[List, str],
        parameter_name: str,
        traces: List,
        show_obs: List,
        highlight_realizations: List[int],
        mode: str,
        parameter_filter: Optional[Dict],
    ) -> Union[Tuple]:
        if not x_column_name:
            raise PreventUpdate
        highlight_realizations = (
            [int(real) for real in highlight_realizations]
            if highlight_realizations
            else []
        )
        if not ensemble_names:
            return [], no_update
        ensemble_names = (
            [ensemble_names] if not isinstance(ensemble_names, list) else ensemble_names
        )
        real_filter = {} if parameter_filter is None else parameter_filter

        csv_dframe = get_table_data(
            tableproviders=tableproviders,
            ensemble_names=ensemble_names,
            table_column_names=[x_column_name, y_column_name],
            realization_filter=real_filter,
        )
        if parameter_name is not None:
            parameter_dframe = get_table_data(
                tableproviders=parameterproviders,
                ensemble_names=ensemble_names,
                table_column_names=[parameter_name],
                realization_filter=real_filter,
            )
            df = pd.merge(csv_dframe, parameter_dframe, on=["ENSEMBLE", "REAL"])
        else:
            df = csv_dframe
        if df.empty:
            return [], {"title": "No data found with current filter"}
        figure = PlotlyLinePlot(
            xaxis_title=x_column_name, yaxis_title=y_column_name, ensemble_colors=colors
        )
        if "Realizations" in traces:
            figure.add_realization_traces(
                dframe=df,
                x_column=x_column_name,
                y_column=y_column_name,
                color_column=parameter_name,
                highlight_reals=highlight_realizations,
                opacity=0.5 if len(traces) > 1 else None,
                mode=mode,
            )
            traces.remove("Realizations")

        if traces:
            stat_df = calc_series_statistics(df, [y_column_name], x_column_name)
            figure.add_statistical_lines(
                dframe=stat_df,
                x_column=x_column_name,
                y_column=y_column_name,
                traces=traces,
                mode=mode,
            )
        if show_obs and observationmodel is not None:

            observations = observationmodel.get_observations_for_attribute(
                attribute=y_column_name, value=x_column_name
            )
            if observations is not None:
                figure.add_observations(
                    observations=observations, x_value=x_column_name
                )

        fig = figure.get_figure()
        data = fig["data"]
        layout = fig["layout"]
        return data, layout

    @app.callback(
        Output(get_uuid("highlight-realizations"), "value"),
        Input(get_uuid("clear-highlight-realizations"), "n_clicks"),
    )
    def _clear_real_highlight(_click: int) -> Optional[List]:
        if _click is not None:
            return []
        raise PreventUpdate

    @app.callback(
        Output(get_uuid("stored_x_value"), "data"),
        Output(get_uuid("traces"), "options"),
        Output(get_uuid("traces"), "value"),
        Output(get_uuid("statistics_warning"), "children"),
        Input(
            {
                "id": get_uuid("data_selectors"),
                "data_attribute": "x",
                "source": "table",
            },
            "value",
        ),
        State(
            {
                "id": get_uuid("data_selectors"),
                "data_attribute": "ensemble",
                "source": "table",
            },
            "value",
        ),
        State(get_uuid("traces"), "value"),
        State({"id": get_uuid("parameter-filter"), "type": "data-store"}, "data"),
    )
    def _update_statistics_options(
        x_column_name: str,
        ensemble_names: List,
        current_values: List,
        parameter_filter: Optional[Dict],
    ) -> Tuple[str, List[Dict], List, str]:
        """Deactivate statistics calculations if x-axis is varying between realizations"""
        if not ensemble_names:
            raise PreventUpdate
        real_filter = {} if parameter_filter is None else parameter_filter
        csv_dframe = get_table_data(
            tableproviders=tableproviders,
            ensemble_names=ensemble_names,
            table_column_names=[x_column_name],
            realization_filter=real_filter,
        )
        for _ens, ens_df in csv_dframe.groupby("ENSEMBLE"):
            realizations = list(ens_df["REAL"].unique())
            equal_x = all(
                (
                    np.all(
                        ens_df.loc[ens_df["REAL"] == real][x_column_name].values
                        == ens_df.loc[ens_df["REAL"] == realizations[0]][
                            x_column_name
                        ].values
                    )
                    for real in realizations
                )
            )
            if not equal_x:
                return (
                    x_column_name,
                    [
                        {"label": "Realizations", "value": "Realizations"},
                        {
                            "label": "Mean",
                            "value": "Mean",
                            "disabled": True,
                        },
                        {
                            "label": "P10/P90",
                            "value": "P10/P90",
                            "disabled": True,
                        },
                        {
                            "label": "Low/High",
                            "value": "Low/High",
                            "disabled": True,
                        },
                    ],
                    ["Realizations"] if "Realizations" in current_values else [],
                    "⚠️ Cannot calculate statistics as x-axis varies between realizations",
                )
        return (
            x_column_name,
            [
                {"label": val, "value": val}
                for val in ["Realizations", "Mean", "P10/P90", "Low/High"]
            ],
            current_values,
            "",
        )


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calc_series_statistics(
    df: pd.DataFrame, vectors: list, refaxis: str = "DATE"
) -> pd.DataFrame:
    """Calculate statistics for given vectors over the ensembles
    refaxis is used if another column than DATE should be used to groupby.
    """
    # Invert p10 and p90 due to oil industry convention.
    def p10(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=90)

    def p90(x: List[float]) -> np.floating:
        return np.nanpercentile(x, q=10)

    # Calculate statistics, ignoring NaNs.
    stat_df = (
        df[["ENSEMBLE", refaxis] + vectors]
        .groupby(["ENSEMBLE", refaxis])
        .agg([np.nanmean, np.nanmin, np.nanmax, p10, p90])
        .reset_index()  # level=["label", refaxis], col_level=0)
    )
    # Rename nanmin, nanmax and nanmean to min, max and mean.
    col_stat_label_map = {
        "nanmin": "min",
        "nanmax": "max",
        "nanmean": "mean",
        "p10": "high_p10",
        "p90": "low_p90",
    }
    stat_df.rename(columns=col_stat_label_map, level=1, inplace=True)

    return stat_df


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def get_table_data(
    tableproviders: Dict[str, EnsembleTableProvider],
    ensemble_names: List,
    table_column_names: List,
    realization_filter: Dict[str, List],
) -> pd.DataFrame:

    dfs = []
    for ens_name in ensemble_names:
        if not realization_filter.get(ens_name):
            dframe = pd.DataFrame(columns=["ENSEMBLE", "REAL"] + table_column_names)
        else:
            provider = tableproviders[ens_name]
            dframe = provider.get_column_data(
                table_column_names, realizations=realization_filter.get(ens_name)
            )
            dframe["ENSEMBLE"] = ens_name
        dfs.append(dframe)
    if len(dfs) > 0:
        return pd.concat(dfs)
    return pd.DataFrame()



########################################
#
#  Copyright (C) 2021-     Equinor ASA
#
#  webviz-subsurface is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  webviz-subsurface is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License at <http:#www.gnu.org/licenses/gpl.html>
#  for more details.
#
########################################
########################################
#
#  The code in this file is based on and/or inspired by opm-common,
#  which is distributed under the GNU General Public License v3.0,
#  and available at https://github.com/OPM/opm-common.
#
#  Especially in the following points this file, or files imported in this file
#  within the same context, make use of or are based on methods developed
#  for the opm-common project:
#  - Reading of Eclipse INIT files
#  - Units, unit systems and conversion of units
#  - Class structures for storing and accessing PVT data,
#    both base classes and derived classes for specific fluids
#
#  You should have received a copy of the GNU General Public License
#  along with webviz-subsurface. If not, see (https://www.gnu.org/licenses/gpl-3.0.html).
#
#  (12-01-2021)
#
########################################

import abc
import warnings
from enum import Enum
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
from scipy import interpolate

# opm is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from opm.io.ecl import EclFile
except ImportError:
    pass

from ..eclipse_unit import ConvertUnits, EclUnitEnum, EclUnits
from ..units import Unit


class EclPropertyTableRawData:  # pylint: disable=too-few-public-methods
    """
    A structure for storing read
    INIT file data.
    """

    def __init__(self) -> None:
        self.data = np.zeros(0)
        self.primary_key: List[int] = []
        self.num_primary = 0
        self.num_rows = 0
        self.num_cols = 0
        self.num_tables = 0


class PvxOBase(abc.ABC):
    """A common base class for all fluids.

    Should be inherited by any new fluid in order to have
    a common interface.
    """

    @abc.abstractmethod
    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def get_independents(self) -> np.ndarray:
        """Returns all independents.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def formation_volume_factor(
        self, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the volume factor values are requested for.
            pressure: Pressure values the volume factor values are requested for.

        Returns:
            All volume factor values for the given ratio and pressure values.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def viscosity(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the viscosity values are requested for.
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            All viscosity values for the given ratio and pressure values.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def density(self, ratio: np.ndarray, pressure: np.ndarray) -> np.ndarray:
        """Args:
            ratio: Ratio (key) values the density values are requested for.
            pressure: Pressure values the density values are requested for.

        Returns:
            All density values for the given ratio and pressure values.

        Base implementation, raises a NotImplementedError.
        """


class PVxx(abc.ABC):
    """A base class for PVDx and PVTx"""

    @abc.abstractmethod
    def get_keys(self) -> np.ndarray:
        """Returns:
            All primary keys.

        Base implementation, raises a NotImplementedError.
        """

    @abc.abstractmethod
    def get_independents(self) -> np.ndarray:
        """Returns:
            All independents.

        Base implementation, raises a NotImplementedError.
        """

    @staticmethod
    def entry_valid(x: float) -> bool:
        """Returns:
        True if the given value is valid, i.e. < 1.0e20, else False.
        """
        return abs(x) < 1.0e20


class PVDx(PVxx):
    """A base class for dead and dry gas/oil respectively.

    Attributes:
        x: The independent values.
        y: A two-dimensional numpy array holding the dependent values.
        interpolation: A scipy interp1d object for interpolating the dependent values.
        inter_extrapolation: An extrap1d object for inter- and extrapolating the dependent values.
    """

    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        convert: ConvertUnits,
    ) -> None:
        """Extracts all values of the table with the given index from raw, converts them according
        to the given convert object and stores them as numpy arrays, x and y respectively.

        Creates an interpolation and an extrapolation object utilising scipy's interp1d and based on
        it the custom tailored extrap1d.

        Raises a ValueError if there is no interpolation interval given, that is when there are
        fewer than two independents.

        Args:
            index_table: The index of the table which values are supposed to be extracted.
            raw:
                An EclPropertyTableRawData object that was initialised based on an Eclipse
                INIT file.
            convert: A ConvertUnit object that contains callables for converting units.

        """
        self.x: np.ndarray = np.zeros(0)
        self.y: np.ndarray = np.zeros((raw.num_cols - 1, 0))

        column_stride = raw.num_rows * raw.num_tables * raw.num_primary
        table_stride = raw.num_primary * raw.num_rows

        for index_primary in range(0, raw.num_primary):
            if self.entry_valid(raw.primary_key[index_primary]):
                for index_row in range(0, raw.num_rows):
                    current_stride = (
                        index_table * table_stride
                        + index_primary * raw.num_rows
                        + index_row
                    )

                    if self.entry_valid(raw.data[column_stride * 0 + current_stride]):
                        self.x = np.append(
                            self.x,
                            convert.independent(
                                raw.data[column_stride * 0 + current_stride]
                            ),
                        )

                        self.y = np.append(
                            self.y,
                            [
                                [
                                    convert.column[index_column - 1](
                                        raw.data[
                                            column_stride * index_column
                                            + current_stride
                                        ]
                                    )
                                ]
                                for index_column in range(1, raw.num_cols)
                            ],
                            axis=1,
                        )

                    else:
                        break
            else:
                break

        if len(self.x) < 2:
            raise ValueError("No interpolation interval of non-zero size.")

        self.__interpolation = interpolate.interp1d(self.x, self.y, axis=1)

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys.

        Since this is dry/dead gas/oil, there is no dependency on Rv/Rs.
        Hence, this method returns a list holding floats of value 0.0
        for each independent value.

        """
        return np.zeros(len(self.x))

    def get_independents(self) -> np.ndarray:
        """Returns all independents.

        In case of gas/oil this returns pressure values.

        """
        return self.x

    def formation_volume_factor(self, pressure: np.ndarray) -> np.ndarray:
        """Computes all formation volume factor values
        for the given pressure values.

        Args:
            pressure: Pressure values the volume factors are requested for.

        Returns:
            All formation volume factor values corresponding
            to the given pressure values.

        """
        # 1 / (1 / B)
        return self.__compute_quantity(pressure, lambda p: 1.0 / self.__fvf_recip(p))

    def viscosity(self, pressure: np.ndarray) -> np.ndarray:
        """Computes all viscosity values for the given pressure values.

        Args:
            pressure: Pressure values the viscosity values are requested for.

        Returns:
            All viscosity values corresponding
            to the given of pressure values.

        """
        # (1 / B) / (1 / (B * mu)
        return self.__compute_quantity(
            pressure, lambda p: self.__fvf_recip(p) / self.__fvf_mu_recip(p)
        )

    @staticmethod
    def __compute_quantity(
        pressures: np.ndarray, evaluate: Callable[[Any], Any]
    ) -> np.ndarray:
        """Calls the given evaluate function with each of the
        given pressure values and returns the results.

        Args:
            pressures: Pressure values
            evaluate: Evaluation function

        Returns:
            Values resulting from evaluating the pressure values.

        """
        result = np.zeros(len(pressures))

        for i, pressure in enumerate(pressures):
            result[i] = float(evaluate(pressure))

        return result

    def __fvf_recip(self, point: float) -> float:
        """Computes (possibly inter-/extrapolates) the reciprocal of
        the formation volume factor for the given point.

        Args:
            point:
                The pressure point the formation volume factor
                is requested for.

        Returns:
            The requested reciprocal formation volume factor.

        """
        return float(self.__interpolation(point)[0])

    def __fvf_mu_recip(self, point: float) -> float:
        """Computes (possibly inter-/extrapolates) the reciprocal of
        the product of the formation volume factor and viscosity
        for the given point.

        Args:
            point:
                The pressure point the formation volume factor
                is requested for.

        Returns:
            The requested reciprocal product of the formation volume factor
            and the viscosity.

        """
        return float(self.__interpolation(point)[1])


class PVTx(PVxx):
    def __init__(
        self,
        index_table: int,
        raw: EclPropertyTableRawData,
        convert: Tuple[
            Callable[
                [
                    float,
                ],
                float,
            ],
            ConvertUnits,
        ],
    ) -> None:
        """Extracts all values of the table with the given index from raw, converts them according
        to the given convert object and stores them as numpy arrays, keys, x and y respectively.

        Creates an interpolation and an extrapolation object utilising scipy's interp2d.

        Args:
            index_table: The index of the table which values are supposed to be extracted.
            raw:
                An EclPropertyTableRawData object that was initialised based on an Eclipse
                INIT file.
            convert:
                A tuple consisting of a callable for converting the primary keys and a
                ConvertUnit object that contains callables for converting units.

        """
        self.keys: np.ndarray = np.zeros(0)
        self.x: np.ndarray = np.zeros(0)
        self.y: List[np.ndarray] = [np.zeros(0) for _ in range(1, raw.num_cols)]

        column_stride = raw.num_rows * raw.num_tables * raw.num_primary
        table_stride = raw.num_primary * raw.num_rows

        for index_primary in range(0, raw.num_primary):
            if self.entry_valid(raw.primary_key[index_primary]):
                for index_row in range(0, raw.num_rows):
                    current_stride = (
                        index_table * table_stride
                        + index_primary * raw.num_rows
                        + index_row
                    )

                    if self.entry_valid(raw.data[column_stride * 0 + current_stride]):
                        self.keys = np.append(
                            self.keys, convert[0](raw.primary_key[index_primary])
                        )
                        self.x = np.append(
                            self.x,
                            convert[1].independent(
                                raw.data[column_stride * 0 + current_stride]
                            ),
                        )

                        for index_column in range(1, raw.num_cols):
                            self.y[index_column - 1] = np.append(
                                self.y[index_column - 1],
                                convert[1].column[index_column - 1](
                                    raw.data[
                                        column_stride * index_column + current_stride
                                    ]
                                ),
                            )

                    else:
                        break
            else:
                break

        # NOTE: If there is only one primary key, interp2d cannot be used.
        # As a fallback, use interp1d and make sure that the primary key asked for in
        # any of the methods of this instance is the one stored in self.keys[0].
        # Extrapolation is not possible.

        self.__single_key = np.amax(self.keys) == np.amin(self.keys)

        if len(self.x) < 2:
            raise ValueError("No interpolation interval of non-zero size.")

        # Ignore warnings here for now since this seems to be a bug in scipy
        # see: https://github.com/scipy/scipy/issues/4138
        # TODO: check if the warning can be fixed.

        warnings.filterwarnings("ignore")

        self.__interpolants: List[Union[interpolate.interp2d, interpolate.interp1d]] = [
            (
                interpolate.interp1d(self.x, self.y[index_column])
                if self.__single_key
                else interpolate.interp2d(
                    x=self.keys, y=self.x, z=self.y[index_column], kind="linear"
                )
            )
            for index_column in range(raw.num_cols - 1)
        ]

        warnings.filterwarnings("default")

    def get_keys(self) -> np.ndarray:
        """Returns all primary keys."""
        return self.keys

    def get_independents(self) -> np.ndarray:
        """Returns all independents."""
        return self.x

    def key_valid(self, key: np.ndarray) -> None:
        if self.__single_key and not all(k == self.keys[0] for k in key):
            raise ValueError(
                "Impossible to perform requested inter-/extrapolation due to insufficient data."
            )

    def formation_volume_factor(self, key: np.ndarray, x: np.ndarray) -> np.ndarray:
        """Computes formation volume factor values
        for the given ratio and pressure values.

        Args:
            key: Primary key values the volume factors are requested for.
            x: Independents the volume factors are requested for.

        Returns:
            Formation volume factor values corresponding
            to the given primary key and independent values.

        Base implementation, raises a NotImplementedError.

        """

        self.key_valid(key)

        return self.__compute_quantity(
            key,
            x,
            lambda curve, point: self.__interpolants[0](point)
            if self.__single_key
            else self.__interpolants[0](curve, point),
            lambda recip_fvf: 1.0 / recip_fvf,
        )

    def viscosity(self, key: np.ndarray, x: np.ndarray) -> np.ndarray:
        """Computes viscosity values for the given ratio and pressure values.

        Args:
            key: Primary key values the viscosity values are requested for.
            x: Independents the viscosity values are requested for.

        Returns:
            Viscosity values corresponding
            to the given primary key and independent values.

        Base implementation, raises a NotImplementedError.

        """

        self.key_valid(key)

        return self.__compute_quantity(
            key,
            x,
            lambda curve, point: [
                self.__interpolants[0](point),
                self.__interpolants[1](point),
            ]
            if self.__single_key
            else [
                self.__interpolants[0](curve, point),
                self.__interpolants[1](curve, point),
            ],
            lambda dense_vector: dense_vector[0] / dense_vector[1],
        )

    @staticmethod
    def __compute_quantity(
        key: np.ndarray,
        x: np.ndarray,
        inner_function: Callable,
        outer_function: Callable,
    ) -> np.ndarray:
        """Calls the evaluate method with each of the values
        in the given primary keys and the given inner_function
        and returns the results after the outer_function
        has been applied on each them.

        Args:
            key: Primary key values the viscosity values are requested for.
            x: Independents the viscosity values are requested for.
            inner_function: Callable for extracting a dense vector.
            outer_function:
                Callable that uses the dense vector to compute
                the requested quantity.

        Returns:
            Result values

        """

        num_vals = len(key)
        results = np.zeros(num_vals)
        if len(x) != num_vals:
            raise ValueError(
                "Number of inner sampling points does not match number of outer sampling points."
            )

        for i in range(0, num_vals):
            quantity = inner_function(key[i], x[i])

            results[i] = float(outer_function(quantity))

        return results


class InitFileDefinitions:  # pylint: disable=too-few-public-methods
    """
    A namespace for constant definitions for
    reading Eclipse INIT files.
    """

    LOGIHEAD_KW = "LOGIHEAD"
    INTEHEAD_KW = "INTEHEAD"
    INTEHEAD_UNIT_INDEX = 2
    INTEHEAD_PHASE_INDEX = 14
    LOGIHEAD_RS_INDEX = 0
    LOGIHEAD_RV_INDEX = 1
    TABDIMS_IBPVTO_OFFSET_ITEM = 6
    TABDIMS_JBPVTO_OFFSET_ITEM = 7
    TABDIMS_NRPVTO_ITEM = 8
    TABDIMS_NPPVTO_ITEM = 9
    TABDIMS_NTPVTO_ITEM = 10
    TABDIMS_IBPVTW_OFFSET_ITEM = 11
    TABDIMS_NTPVTW_ITEM = 12
    TABDIMS_IBPVTG_OFFSET_ITEM = 13
    TABDIMS_JBPVTG_OFFSET_ITEM = 14
    TABDIMS_NRPVTG_ITEM = 15
    TABDIMS_NPPVTG_ITEM = 16
    TABDIMS_NTPVTG_ITEM = 17
    LOGIHEAD_CONSTANT_OILCOMPR_INDEX = 39 - 1

    TABDIMS_IBDENS_OFFSET_ITEM = 18
    TABDIMS_NTDENS_ITEM = 19


class EclPhaseIndex(Enum):
    """Enumerator holding the different phases according
    to Eclipse file conventions"""

    AQUA = 0
    LIQUID = 1
    VAPOUR = 2


def is_const_compr_index() -> int:
    """Convenient function for better readibility.

    Returns:
        An integer that states that the oil has constant compression
        according to Eclipse LOGIHEAD file conventions.
    """
    return InitFileDefinitions.LOGIHEAD_CONSTANT_OILCOMPR_INDEX


def surface_mass_density(
    ecl_file: EclFile, phase: EclPhaseIndex, keep_unit_system: bool = True
) -> np.ndarray:
    """Extracts the surface mass density from the given Eclipse file for the given phase.

    Args:
        ecl_file: The Eclipse file to extract data from
        phase: Fluid phase to extract data for

    Returns:
        Surface mass density values.

    """
    if phase is EclPhaseIndex.LIQUID:
        col = 0
    elif phase is EclPhaseIndex.AQUA:
        col = 1
    elif phase is EclPhaseIndex.VAPOUR:
        col = 2
    else:
        raise AttributeError("Phase must be Liquid, Water or Vapour.")

    tabdims = ecl_file.__getitem__("TABDIMS")
    tab = ecl_file.__getitem__("TAB")

    start = tabdims[InitFileDefinitions.TABDIMS_IBDENS_OFFSET_ITEM] - 1
    nreg = tabdims[InitFileDefinitions.TABDIMS_NTDENS_ITEM]

    rho = tab[start + nreg * (col + 0) : start + nreg * (col + 1)]

    intehead = ecl_file.__getitem__(InitFileDefinitions.INTEHEAD_KW)
    unit_system = EclUnits.create_unit_system(
        intehead[InitFileDefinitions.INTEHEAD_UNIT_INDEX]
    )

    if not keep_unit_system:
        rho = [Unit.Convert.from_(rho_i, unit_system.density().value) for rho_i in rho]

    return rho


class FluidImplementation(abc.ABC):
    """Base class for fluid implementations

    Holds a list of regions (one per PVT table).

    Attributes:
        keep_unit_system: True if the original unit system was kept
        original_unit_system: An ErtEclUnitEnum representing the original unit system
    """

    class InvalidArgument(Exception):
        """An exception for invalid arguments"""

        def __init__(self, message: str):
            self.message = message
            super().__init__(message)

    class InvalidType(Exception):
        """An exception for invalid types"""

        def __init__(self) -> None:
            super().__init__("Invalid type. Only live oil/wet gas/water supported.")

    def __init__(self, unit_system: int, keep_unit_system: bool = True) -> None:
        """Initializes a fluid object.

        Args:
            unit_system: The original unit system
            keep_unit_system:
                True if the original unit system shall be kept,
                False if units shall be converted to SI units.

        """
        self._regions: List[PvxOBase] = []
        self.keep_unit_system = keep_unit_system
        self.original_unit_system = unit_system

    def pvdx_unit_converter(self) -> Optional[ConvertUnits]:
        """Creates a pseudo ConvertUnits object for PVDx interpolants
        that keeps the old unit system.

        Returns:
            Pseudo ConvertUnits object that does not do any unit conversions.

        """
        if self.keep_unit_system:
            return ConvertUnits(
                lambda x: x,
                [
                    lambda x: x,
                    lambda x: x,
                    lambda x: x,
                    lambda x: x,
                ],
            )
        return None

    def pvtx_unit_converter(
        self,
    ) -> Optional[Tuple[Callable[[float,], float,], ConvertUnits]]:
        """Creates a tuple consisting of a callable and a pseudo ConvertUnits
        object for PVTx interpolants which both keep the old unit system.

        Returns:
            Tuple of callable and pseudo ConvertUnits object
            which do not do any unit conversions.

        """
        if self.keep_unit_system:
            return (
                lambda x: x,
                ConvertUnits(
                    lambda x: x,
                    [
                        lambda x: x,
                        lambda x: x,
                        lambda x: x,
                        lambda x: x,
                    ],
                ),
            )
        return None

    @staticmethod
    def make_interpolants_from_raw_data(
        raw: EclPropertyTableRawData,
        construct: Callable[[int, EclPropertyTableRawData], PvxOBase],
    ) -> List[PvxOBase]:
        """Creates a list of interpolants from raw Eclipse table data using
        the given construct callable.

        Args:
            raw: Raw Eclipse table data to create interpolants from
            construct: Callable to use when creating interpolants

        Returns:
            A list of created interpolants.
        """
        interpolants: List[PvxOBase] = []

        for table_index in range(0, raw.num_tables):
            interpolants.append(construct(table_index, raw))

        return interpolants

    def pressure_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as Latex, False if not.

        Returns:
            A string containing the unit symbol of pressure.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return rf"${unit_system.pressure().symbol}$"
        return f"{unit_system.pressure().symbol}"

    def formation_volume_factor(
        self, region_index: int, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            region_index: Index of the requested PVT region
            ratio: Ratio values the data is requested for
            pressure: Pressure values the data is requested for

        Returns:
            Formation volume factor values according to the given values.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index].formation_volume_factor(ratio, pressure)

    @abc.abstractmethod
    def formation_volume_factor_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the formation volume factor.

        Raises a NotImplementedError when called on a base class object.

        """

    def viscosity(
        self, region_index: int, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            region_index: Index of the requested PVT region
            ratio: Ratio values the data is requested for
            pressure: Pressure values the data is requested for

        Returns:
            Viscosity values according to the given values.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index].viscosity(ratio, pressure)

    def viscosity_unit(self, latex: bool = False) -> str:
        """Creates and returns the unit symbol of the viscosity.

        Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the viscosity.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return rf"${unit_system.viscosity().symbol}$"
        return unit_system.viscosity().symbol

    def density(
        self, region_index: int, ratio: np.ndarray, pressure: np.ndarray
    ) -> np.ndarray:
        """Args:
            region_index: Index of the requested PVT region
            ratio: Ratio values the data is requested for
            pressure: Pressure values the data is requested for

        Returns:
            Density values according to the given values.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index].density(ratio, pressure)

    def density_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            The unit symbol of the density.

        """
        unit_system = EclUnits.create_unit_system(
            self.original_unit_system
            if self.keep_unit_system
            else EclUnitEnum.ECL_SI_UNITS
        )

        if latex:
            return rf"${unit_system.density().symbol}$"
        return unit_system.density().symbol

    @abc.abstractmethod
    def ratio_unit(self, latex: bool = False) -> str:
        """Args:
            latex: True if the unit symbol shall be returned as LaTeX, False if not.

        Returns:
            A string containing the unit symbol of the phase ratio (e.g. Sm³/Sm³).

        Raises a NotImplementedError when called on a base class object.

        """

    def get_region(self, region_index: int) -> PvxOBase:
        """Validates and returns the region at the given region_index.

        Args:
            region_index: Index of the requested PVT region

        Returns:
            The fluid interpolant related to the given region index.

        """
        self.validate_region_index(region_index)

        return self._regions[region_index]

    def validate_region_index(self, region_index: int) -> None:
        """Validates the given region index by ensuring that it is not out of range.

        Args:
            region_index: Index of the region to validate

        Raises:
            TypeError if the region is invalid.

        """
        if region_index >= len(self.regions()):
            if len(self.regions()) == 0:
                raise TypeError(
                    f"No oil PVT interpolant available in region {region_index + 1}."
                )

            raise TypeError(
                f"Region index {region_index + 1} outside valid range 0"
                f", ..., {len(self._regions) - 1}."
            )

    def regions(self) -> List[PvxOBase]:
        """Returns:
        All interpolants (one interpolant per region)

        """
        return self._regions

    def range_ratio(self, region_index: int) -> Tuple[float, float]:
        """Gets the primary key (ratio) range of the PVT region with the given index.

        Args:
            region_index: Index of the requested PVT region

        Returns:
            A tuple containing the min and max primary key (ratio) values.

        """
        region = self.get_region(region_index)
        return (min(region.get_keys()), max(region.get_keys()))

    def range_independent(self, region_index: int) -> Tuple[float, float]:
        """Gets the independent variable range of the PVT region with the given index.

        Args:
            region_index: Index of the requested PVT region

        Returns:
            A tuple containing the min and max independent values.

        """
        region = self.get_region(region_index)
        return (min(region.get_independents()), max(region.get_independents()))



from typing import Callable, List

import webviz_core_components as wcc
from dash import html


def line_traces_view(
    get_uuid: Callable,
) -> html.Div:

    return wcc.Selectors(
        label="Plot data",
        children=[
            wcc.RadioItems(
                id=get_uuid("mode"),
                options=[
                    {"value": "lines", "label": "Lines"},
                    {"value": "markers", "label": "Points"},
                ],
                value="lines",
            ),
            wcc.Checklist(
                id=get_uuid("traces"),
                options=[
                    {"label": val, "value": val}
                    for val in ["Realizations", "Mean", "P10/P90", "Low/High"]
                ],
                value=["Realizations"],
                labelStyle={"display": "block"},
            ),
            wcc.Checklist(
                id=get_uuid("observations"),
                options=[{"label": val, "value": val} for val in ["Observations"]],
                value=["Observations"],
                labelStyle={"display": "block"},
            ),
            wcc.Label(id=get_uuid("statistics_warning"), children=""),
        ],
    )


def highlight_realizations_view(
    get_uuid: Callable, realizations: List[int]
) -> html.Div:

    return wcc.Selectors(
        label="Highlight realizations",
        children=[
            wcc.Select(
                id=get_uuid("highlight-realizations"),
                options=[{"label": val, "value": val} for val in realizations],
                value=None,
            ),
            html.Button(
                id=get_uuid("clear-highlight-realizations"),
                children="Clear",
            ),
        ],
    )



from webviz_subsurface._providers import VectorMetadata

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.trace_line_shape import (
    get_simulation_line_shape,
)


def test_get_simulation_line_shape() -> None:
    total_vector_metadata = VectorMetadata(
        unit="M3",
        is_total=True,
        is_rate=False,
        is_historical=False,
        keyword="Test",
        wgname=None,
        get_num=None,
    )

    rate_vector_metadata = VectorMetadata(
        unit="M3/Day",
        is_total=False,
        is_rate=True,
        is_historical=False,
        keyword="Test rate",
        wgname=None,
        get_num=None,
    )

    fallthrough_vector_metadata = VectorMetadata(
        unit="M3/M3",
        is_total=False,
        is_rate=False,
        is_historical=False,
        keyword="Test fallthrough",
        wgname=None,
        get_num=None,
    )

    assert get_simulation_line_shape("Fallback", "PER_INTVL_vector", None) == "hv"
    assert get_simulation_line_shape("Fallback", "PER_DAY_vector", None) == "hv"
    assert get_simulation_line_shape("Fallback", "test_vector", None) == "Fallback"
    assert (
        get_simulation_line_shape("Fallback", "test_vector", total_vector_metadata)
        == "linear"
    )
    assert (
        get_simulation_line_shape("Fallback", "test_vector", rate_vector_metadata)
        == "vh"
    )
    assert (
        get_simulation_line_shape(
            "Fallback", "test_vector", fallthrough_vector_metadata
        )
        == "Fallback"
    )



from typing import Callable, Dict, List, Optional, Tuple

from dash import Dash, Input, Output, State, callback_context
from dash.exceptions import PreventUpdate


def update_realizations(app: Dash, get_uuid: Callable) -> None:
    @app.callback(
        Output(
            {
                "id": get_uuid("dialog"),
                "dialog_id": "realization-filter",
                "element": "apply",
            },
            "disabled",
        ),
        Output(
            {
                "id": get_uuid("dialog"),
                "dialog_id": "realization-filter",
                "element": "apply",
            },
            "style",
        ),
        Input(
            get_uuid("realization-store"),
            "data",
        ),
        Input(
            {"id": get_uuid("intersection-data"), "element": "realizations"},
            "value",
        ),
    )
    def _activate_realization_apply_btn(
        stored_reals: List, selected_reals: List
    ) -> Tuple[bool, Dict[str, str]]:
        if stored_reals is None or selected_reals is None:
            raise PreventUpdate
        if set(stored_reals) == set(selected_reals):
            return True, {"visibility": "hidden"}
        return False, {"visibility": "visible"}

    @app.callback(
        Output(
            get_uuid("realization-store"),
            "data",
        ),
        Input(
            {
                "id": get_uuid("dialog"),
                "dialog_id": "realization-filter",
                "element": "apply",
            },
            "n_clicks",
        ),
        State(
            {"id": get_uuid("intersection-data"), "element": "realizations"},
            "value",
        ),
    )
    def _store_realizations(btn_click: Optional[int], selected_reals: List) -> List:
        if btn_click:
            return selected_reals
        raise PreventUpdate

    @app.callback(
        Output(
            {"id": get_uuid("intersection-data"), "element": "realizations"},
            "value",
        ),
        Input(
            {
                "id": get_uuid("dialog"),
                "dialog_id": "realization-filter",
                "element": "clear",
            },
            "n_clicks",
        ),
        Input(
            {
                "id": get_uuid("dialog"),
                "dialog_id": "realization-filter",
                "element": "all",
            },
            "n_clicks",
        ),
        State(
            {"id": get_uuid("intersection-data"), "element": "realizations"},
            "options",
        ),
    )
    def _update_realization_list(
        clear_click: Optional[int], all_click: Optional[int], real_opts: List
    ) -> List:
        if clear_click is None and all_click is None:
            raise PreventUpdate
        ctx = callback_context.triggered
        if "clear" in ctx[0]["prop_id"]:
            return []
        if "all" in ctx[0]["prop_id"]:
            return [opt["value"] for opt in real_opts]
        raise PreventUpdate



from typing import Any, List, Tuple

import numpy as np
import pandas as pd
import plotly.graph_objects as go

from webviz_subsurface._figures import create_figure
from webviz_subsurface._models.parameter_model import ParametersModel as Pmodel


class ParametersModel:
    """Class to process and visualize ensemble parameter data"""

    REQUIRED_COLUMNS = ["ENSEMBLE", "REAL"]

    def __init__(
        self, dataframe: pd.DataFrame, theme: dict, drop_constants: bool = True
    ) -> None:
        self.pmodel = Pmodel(
            dataframe=dataframe, drop_constants=drop_constants, keep_numeric_only=True
        )
        self._dataframe = self.pmodel.dataframe
        self._dataframe["REAL"] = self._dataframe["REAL"].astype(int)
        self._parameters = self.pmodel.parameters
        self.theme = theme
        self.colorway = self.theme.plotly_theme.get("layout", {}).get("colorway", None)
        self._statframe = self._aggregate_ensemble_data(self._dataframe)
        self._statframe_normalized = self._normalize_and_aggregate()
        self._dataframe_melted = self.dataframe.melt(
            id_vars=["ENSEMBLE", "REAL"], var_name="PARAMETER", value_name="VALUE"
        )

    @property
    def dataframe_melted(self) -> pd.DataFrame:
        return self._dataframe_melted

    @property
    def dataframe(self) -> pd.DataFrame:
        return self._dataframe

    @property
    def statframe(self) -> pd.DataFrame:
        return self._statframe

    @property
    def mc_ensembles(self) -> pd.DataFrame:
        return self.pmodel.mc_ensembles

    @property
    def parameters(self) -> pd.DataFrame:
        return self._parameters

    @parameters.setter
    def parameters(self, sortorder):
        self._parameters = sortorder

    @property
    def ensembles(self) -> List[str]:
        return list(self.dataframe["ENSEMBLE"].unique())

    @staticmethod
    def _aggregate_ensemble_data(dframe) -> pd.DataFrame:
        """Compute parameter statistics for the different ensembles"""
        return (
            dframe.drop(columns=["REAL"])
            .groupby(["ENSEMBLE"])
            .agg(
                [
                    ("Avg", np.mean),
                    ("Stddev", np.std),
                    ("P10", lambda x: np.percentile(x, 10)),
                    ("P90", lambda x: np.percentile(x, 90)),
                    ("Min", np.min),
                    ("Max", np.max),
                ]
            )
            .stack(0)
            .rename_axis(["ENSEMBLE", "PARAMETER"])
            .reset_index()
        )

    def _normalize_and_aggregate(self):
        """
        Normalize parameter values to be able to compare distribution updates
        for different parameters
        """
        df = self._dataframe.copy()
        df_norm = (df[self.parameters] - df[self.parameters].min()) / (
            df[self.parameters].max() - df[self.parameters].min()
        )
        df_norm[self.REQUIRED_COLUMNS] = df[self.REQUIRED_COLUMNS]
        df = self._aggregate_ensemble_data(df_norm)
        return df.pivot_table(columns=["ENSEMBLE"], index="PARAMETER").reset_index()

    def sort_parameters(
        self,
        ensemble: str,
        delta_ensemble: str,
        sortby: str,
    ):
        """Sort parameter list from selection"""
        # compute diff between ensembles
        df = self._statframe_normalized.copy()
        df["Avg", "diff"] = abs(df["Avg"][ensemble] - df["Avg"][delta_ensemble])
        df["Stddev", "diff"] = df["Stddev"][ensemble] - df["Stddev"][delta_ensemble]

        # set parameter column and update parameter list
        df = df.sort_values(
            by="PARAMETER" if sortby == "Name" else [(sortby, "diff")],
            ascending=(sortby == "Name"),
        )
        self._parameters = list(df["PARAMETER"])
        return list(df["PARAMETER"])

    @staticmethod
    def make_table(df: pd.DataFrame) -> Tuple[List[Any], List[Any]]:
        """Return format needed for dash table"""
        col_order = ["PARAMETER", "Avg", "Stddev", "P90", "P10", "Min", "Max"]
        df = df.reindex(col_order, axis=1, level=0)
        df.columns = df.columns.map("|".join)
        columns = [
            {
                "id": col,
                "name": [col.split("|")[0], col.split("|")[1]],
                "type": "numeric",
                "format": {"specifier": ".5~r"},
            }
            for col in df.columns
        ]
        return columns, df.to_dict("records")

    def _sort_parameters_col(self, df, parameters):
        """Sort parameter column in dataframe"""
        sortorder = [x for x in self._parameters if x in parameters]
        return df.set_index("PARAMETER").loc[sortorder].reset_index()

    def make_statistics_table(
        self,
        ensembles: list,
        parameters: List[Any],
    ) -> Tuple[List[Any], List[Any]]:
        """Create table with statistics for selected parameters"""
        df = self.statframe.copy()
        df = df[df["ENSEMBLE"].isin(ensembles)]
        df = df[df["PARAMETER"].isin(parameters)]
        df = df.pivot_table(columns=["ENSEMBLE"], index="PARAMETER").reset_index()
        df = self._sort_parameters_col(df, parameters)
        return self.make_table(df)

    def make_grouped_plot(
        self,
        ensembles: list,
        parameters: List[Any],
        plot_type: str = "distribution",
    ) -> go.Figure:
        """Create subplots for selected parameters"""
        df = self.dataframe_melted.copy()
        df = df[df["ENSEMBLE"].isin(ensembles)]
        df = df[df["PARAMETER"].isin(parameters)]
        df = self._sort_parameters_col(df, parameters)

        return (
            create_figure(
                plot_type=plot_type,
                data_frame=df,
                x="VALUE",
                facet_col="PARAMETER",
                color="ENSEMBLE",
                color_discrete_sequence=self.colorway,
            )
            .update_xaxes(matches=None)
            .for_each_trace(
                lambda t: t.update(
                    text=t["text"].replace("VALUE", "")
                    if t["text"] is not None
                    else None
                )
            )
        )

    def get_stat_value(self, parameter: str, ensemble: str, stat_column: str):
        """
        Retrive statistical value for a parameter in an ensamble.
        """
        return self.statframe.loc[
            (self.statframe["PARAMETER"] == parameter)
            & (self.statframe["ENSEMBLE"] == ensemble)
        ].iloc[0][stat_column]

    def get_real_and_value_df(
        self, ensemble: str, parameter: str, normalize: bool = False
    ) -> pd.DataFrame:
        """
        Return dataframe with ralization and values for selected parameter for an ensemble.
        A column with normalized parameter values can be added.
        """
        df = self.dataframe_melted.copy()
        df = df[["VALUE", "REAL"]].loc[
            (df["ENSEMBLE"] == ensemble) & (df["PARAMETER"] == parameter)
        ]
        if normalize:
            df["VALUE_NORM"] = (df["VALUE"] - df["VALUE"].min()) / (
                df["VALUE"].max() - df["VALUE"].min()
            )
        return df.reset_index(drop=True)

    def get_parameter_df_for_ensemble(self, ensemble: str, reals: list):
        return self._dataframe[
            (self._dataframe["ENSEMBLE"] == ensemble)
            & (self._dataframe["REAL"].isin(reals))
        ]



import math
from typing import List, Optional, Union

import plotly.graph_objects as go
import webviz_core_components as wcc
from dash import dash_table


def create_table_columns(
    columns: list,
    text_columns: list = None,
    use_si_format: Optional[list] = None,
    use_percentage: Optional[list] = None,
) -> List[dict]:

    text_columns = text_columns if text_columns is not None else []
    use_si_format = use_si_format if use_si_format is not None else []
    use_percentage = use_percentage if use_percentage is not None else []

    table_columns = []
    for col in columns:
        data = {"id": col, "name": col}
        if col not in text_columns:
            data["type"] = "numeric"
            if col in use_percentage:
                data["format"] = {"specifier": ".1f"}
            elif col in use_si_format:
                data["format"] = {"locale": {"symbol": ["", ""]}, "specifier": "$.4s"}
            else:
                data["format"] = {"specifier": ".3~r"}
        table_columns.append(data)
    return table_columns


def create_data_table(
    columns: list,
    height: str,
    data: List[dict],
    table_id: dict,
    selectors: Optional[list] = None,
    style_cell: Optional[dict] = None,
    style_cell_conditional: Optional[list] = None,
    style_data_conditional: Optional[list] = None,
) -> Union[list, wcc.WebvizPluginPlaceholder]:

    if not data:
        return []

    if selectors is None:
        selectors = []
    conditional_cell_style = [
        {
            "if": {"column_id": selectors + ["Response", "Property", "Sensitivity"]},
            "width": "10%",
            "textAlign": "left",
        },
        {"if": {"column_id": "FLUID_ZONE"}, "width": "10%", "textAlign": "right"},
    ]
    if style_cell_conditional is not None:
        conditional_cell_style.extend(style_cell_conditional)

    style_data_conditional = (
        style_data_conditional if style_data_conditional is not None else []
    )
    style_data_conditional.extend(fluid_table_style())

    return wcc.WebvizPluginPlaceholder(
        id={"request": "table_data", "table_id": table_id["table_id"]},
        buttons=["expand", "download"],
        children=dash_table.DataTable(
            id=table_id,
            sort_action="native",
            sort_mode="multi",
            filter_action="native",
            columns=columns,
            data=data,
            style_as_list_view=True,
            style_cell=style_cell,
            style_cell_conditional=conditional_cell_style,
            style_data_conditional=style_data_conditional,
            style_table={
                "height": height,
                "overflowY": "auto",
            },
        ),
    )


def fluid_table_style() -> list:
    fluid_colors = {
        "oil": "#007079",
        "gas": "#FF1243",
        "water": "#ADD8E6",
    }
    return [
        {
            "if": {
                "filter_query": "{FLUID_ZONE} = " + f"'{fluid}'",
                "column_id": "FLUID_ZONE",
            },
            "color": color,
            "fontWeight": "bold",
        }
        for fluid, color in fluid_colors.items()
    ]


def fluid_annotation(selections: dict) -> dict:
    fluid_text = (" + ").join(selections["filters"]["FLUID_ZONE"])
    return dict(
        visible=bool(selections["Fluid annotation"])
        and selections["Subplots"] != "FLUID_ZONE",
        x=1,
        y=1,
        xref="paper",
        yref="paper",
        showarrow=False,
        text="Fluid zone<br>" + fluid_text,
        font=dict(size=15, color="black"),
        bgcolor="#E8E8E8",
    )


def add_correlation_line(figure: go.Figure, xy_min: float, xy_max: float) -> go.Figure:
    return figure.add_shape(
        type="line",
        layer="below",
        xref="x",
        yref="y",
        x0=xy_min,
        y0=xy_min,
        x1=xy_max,
        y1=xy_max,
        line=dict(color="black", width=2, dash="dash"),
    )


def create_figure_matrix(figures: List[go.Figure]) -> List[List[go.Figure]]:
    """Convert a list of figures into a matrix for display"""
    figs_in_row = min(
        min([x for x in range(100) if (x * (x + 1)) > len(figures)]),
        20,
    )
    len_of_matrix = figs_in_row * math.ceil(len(figures) / figs_in_row)
    # extend figure list with None to fit size of matrix
    figures.extend([None] * (len_of_matrix - len(figures)))
    return [figures[i : i + figs_in_row] for i in range(0, len_of_matrix, figs_in_row)]


def update_tornado_figures_xaxis(figures: List[go.Figure]) -> None:
    """
    Update the x-axis range for a list of tornado figures with the maximum absolute
    x-value from all figures. Axis will be centered around 0.
    """
    x_absmax = max([max(abs(trace.x)) for fig in figures for trace in fig.data])
    for fig in figures:
        fig.update_layout(xaxis_range=[-x_absmax, x_absmax])



# pylint: disable=too-many-lines
import warnings
from pathlib import Path
from typing import Optional, Union

import numpy as np
import webviz_core_components as wcc
from dash import Input, Output, State, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_assets import WEBVIZ_ASSETS

import webviz_subsurface

from .._datainput.fmu_input import load_csv
from .._datainput.relative_permeability import load_satfunc, load_scal_recommendation
from .._utils.colors import StandardColors
from .._utils.fanchart_plotting import (
    FanchartData,
    FreeLineData,
    LowHighData,
    MinMaxData,
    get_fanchart_traces,
)


class RelativePermeability(WebvizPluginABC):
    """Visualizes relative permeability and capillary pressure curves for FMU ensembles.

---

* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`relpermfile`:** Local path to a csvfile in each realization with dumped relperm data.
* **`scalfile`:** Path to a reference file with SCAL recommendationed data. \
    Path to a single file, **not** per realization/ensemble. The path can be absolute or \
    relative to the `webviz` configuration.
* **`sheet_name`:** Which sheet to use for the `scalfile`, only relevant if `scalfile` is an \
    `xlsx` file (recommended to use csv files with `webviz`).
* **`scal_scenarios`:** Alternative to ensemble based input, using only a dict of where the \
    keys are arbitrary names used for each scenario and the values are absolute path to \
    pyscal files for each scenario (if using `scal_scenarios`, this has to be the **only** \
    input option used).

---
The minimum requirement is to define `ensembles`.

If no `relpermfile` is defined, the relative permeability data will be extracted automatically
from the simulation decks of individual realizations using `fmu-ensemble`and `ecl2df` behind the
scenes. Note that this method can be very slow for larger data decks, and is therefore not
recommended unless you have a very simple model/data deck.

`relpermfile` is a path to a file stored per realization (e.g. in \
`share/results/tables/relperm.csv`). `relpermfile` columns:
* One column named `KEYWORD` or `TYPE`: with Flow/Eclipse style keywords (e.g. `SWOF` and `SGOF`).
* One column named `SATNUM` with integer `SATNUM` regions.
* One column **per** saturation (e.g. `SG` and `SW`).
* One column **per** relative permeability curve (e.g. `KRW`, `KROW` and `KRG`)
* One column **per** capillary pressure curve (e.g. `PCOW`).

The `relpermfile` file can e.g. be dumped to disk per realization by a forward model in ERT that
wraps the command `ecl2csv satfunc input_file -o output_file` (requires that you have `ecl2df`
installed). A typical example could be:
`ecl2csv satfunc eclipse/include/props/relperm.inc -o share/results/tables/relperm.csv`.
[Link to ecl2csv satfunc documentation.](https://equinor.github.io/ecl2df/scripts.html#satfunc)


`scalfile` is a path to __a single file of SCAL recommendations__ (for all
realizations/ensembles). The file has to be compatible with
[pyscal's](https://equinor.github.io/pyscal/pyscal.html#pyscal.\
factory.PyscalFactory.load_relperm_df) input format. Including this file adds reference cases
`Pess`, `Base` and `Opt` to the plots. This file is typically a result of a SCAL study.

`sheet_name` defines the sheet to use in the `scalfile`. Only relevant if `scalfile` is an
`xlsx` file (it is recommended to use `csv` and not `xlsx` for `Webviz`).

`scal_scenarios` is a dict with pyscal formatted files (similar to `scalfile`) as values,
and arbitrary names for each scenario as keys.

* [Example of relpermfile](https://github.com/equinor/webviz-subsurface-testdata/blob/master/\
reek_history_match/realization-0/iter-0/share/results/tables/relperm.csv).
* [Example of scalfile](https://github.com/equinor/\
webviz-subsurface-testdata/blob/master/reek_history_match/share/scal/scalreek.csv).
"""

    SATURATIONS = ["SW", "SO", "SG", "SL"]
    RELPERM_FAMILIES = {
        1: ["SWOF", "SGOF", "SLGOF"],
        2: ["SWFN", "SGFN", "SOF3"],
    }
    SCAL_COLORMAP = {
        "Missing": "#ffff00",  # using yellow if the curve could not be found
        "KRW": StandardColors.WATER_BLUE,
        "KRG": StandardColors.GAS_RED,
        "KROG": StandardColors.OIL_GREEN,
        "KROW": StandardColors.OIL_GREEN,
        "PCOW": "#555555",  # Reserving #000000 for reference envelope (scal rec)
        "PCOG": "#555555",
    }
    # pylint: disable=too-many-branches,too-many-statements
    def __init__(
        self,
        app,
        webviz_settings: WebvizSettings,
        ensembles: Optional[list] = None,
        relpermfile: str = None,
        scalfile: Path = None,
        sheet_name: Optional[Union[str, int, list]] = None,
        scal_scenarios: Optional[dict] = None,
    ):

        super().__init__()

        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "css"
            / "block_options.css"
        )
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.scalfile = scalfile
        self.sheet_name = sheet_name
        self.scal_scenarios = scal_scenarios
        if scal_scenarios:
            self.plugin_mode = "scal_scenarios"
            if any((ensembles, relpermfile, scalfile, sheet_name)):
                raise ValueError(
                    "`scal_scenarios` can only be used alone as an argument."
                )
            self.ensembles = []
            self.scal_ensembles = {}
            self.sat_axes_maps = {}
            satnums = set()

            for case, path in scal_scenarios.items():
                self.ensembles.append(case)
                self.scal_ensembles[case] = load_scal_recommendation(Path(path))
                satnums.update(self.scal_ensembles[case]["SATNUM"].unique())
                for sat_axis in RelativePermeability.SATURATIONS:
                    if sat_axis in self.scal_ensembles[case]:
                        if sat_axis not in self.sat_axes_maps:
                            self.sat_axes_maps[sat_axis] = set()
                        if sat_axis == "SW":
                            self.sat_axes_maps[sat_axis].update(
                                [
                                    col
                                    for col in self.scal_ensembles[case].columns
                                    if col in ["KRW", "KROW", "PCOW"]
                                ]
                            )
                        if sat_axis == "SG":
                            self.sat_axes_maps[sat_axis].update(
                                [
                                    col
                                    for col in self.scal_ensembles[case].columns
                                    if col in ["KRG", "KROG", "PCOG"]
                                ]
                            )
                        if sat_axis == "SL":
                            self.sat_axes_maps[sat_axis].update(
                                [
                                    col
                                    for col in self.scal_ensembles[case].columns
                                    if col in ["KRG", "KROG", "PCOG"]
                                ]
                            )
                        if sat_axis == "SO":
                            self.sat_axes_maps[sat_axis].update(
                                [
                                    col
                                    for col in self.scal_ensembles[case].columns
                                    if col in ["KROW", "KROG"]
                                ]
                            )
            self.sat_axes_maps = {
                sat_axis: list(cols) for sat_axis, cols in self.sat_axes_maps.items()
            }
            self.satnums = sorted(satnums)
        else:
            self.plugin_mode = "distribution"
            self.ens_paths = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.ensembles = ensembles
            self.relpermfile = relpermfile
            if self.relpermfile is not None:
                self.satfunc = load_csv(
                    ensemble_paths=self.ens_paths, csv_file=relpermfile
                )
                self.satfunc = self.satfunc.rename(str.upper, axis="columns").rename(
                    columns={"TYPE": "KEYWORD"}
                )
                if "KEYWORD" not in self.satfunc.columns:
                    raise ValueError(
                        "There has to be a KEYWORD or TYPE column with corresponding Eclipse "
                        "keyword: e.g SWOF, SGOF and etc."
                    )
                valid_columns = (
                    ["ENSEMBLE", "REAL", "KEYWORD", "SATNUM"]
                    + RelativePermeability.SATURATIONS
                    + [
                        key
                        for key in RelativePermeability.SCAL_COLORMAP
                        if key != "Missing"
                    ]
                )
                self.satfunc = self.satfunc[
                    [col for col in self.satfunc.columns if col in valid_columns]
                ]
            else:
                self.satfunc = load_satfunc(self.ens_paths)
            if any(
                keyword in RelativePermeability.RELPERM_FAMILIES[1]
                for keyword in self.satfunc["KEYWORD"].unique()
            ):
                self.family = 1
                if any(
                    keyword in RelativePermeability.RELPERM_FAMILIES[2]
                    for keyword in self.satfunc["KEYWORD"].unique()
                ):
                    warnings.warn(
                        (
                            "Mix of keyword family 1 and 2, currently only support one family at "
                            "the time. Dropping all data of family 2 ('SWFN', 'SGFN', 'SGWFN', "
                            "'SOF2', 'SOF3', 'SOF32D') and continues with family 1 ('SWOF', "
                            "'SGOF', 'SLGOF')."
                        ),
                    )
                    self.satfunc = self.satfunc[
                        self.satfunc["KEYWORD"].isin(
                            RelativePermeability.RELPERM_FAMILIES["fam1"]
                        )
                    ]
                if "SGOF" in self.satfunc["KEYWORD"].unique():
                    if "SLGOF" in self.satfunc["KEYWORD"].unique():
                        warnings.warn(
                            (
                                "Mix of 'SGOF' and 'SLGOF' in ensembles, resulting in non-unique "
                                "horizontal axis ('SG' and 'SL') for 'KRG', 'KROG' and 'PCOG'. "
                                "Dropping all data with 'SLGOF'."
                            ),
                        )
                        self.satfunc = self.satfunc[self.satfunc["KEYWORD"] != "SLGOF"]
                    self.sat_axes_maps = {
                        "SW": ["KRW", "KROW", "PCOW"],
                        "SG": ["KRG", "KROG", "PCOG"],
                    }
                else:
                    self.sat_axes_maps = {
                        "SW": ["KRW", "KROW", "PCOW"],
                        "SL": ["KRG", "KROG", "PCOG"],
                    }
            elif not all(
                keyword in RelativePermeability.RELPERM_FAMILIES[2]
                for keyword in self.satfunc["KEYWORD"].unique()
            ):
                raise ValueError(
                    "Unrecognized saturation table keyword in data. This should not occur unless "
                    "there has been changes to ecl2df. Update of this plugin might be required."
                )
            else:
                self.family = 2
                self.sat_axes_maps = {
                    "SW": ["KRW", "PCOW"],
                    "SG": ["KRG", "PCOG"],
                    "SO": ["KROW", "KROG"],
                }
            self.scal_ensembles = (
                {"SCAL": load_scal_recommendation(self.scalfile, self.sheet_name)}
                if self.scalfile is not None
                else None
            )
            self.satnums = sorted(self.satfunc["SATNUM"].unique())

        self.set_callbacks(app)

    @property
    def sat_axes(self):
        """List of all possible saturation axes in dataframe"""
        return (
            list(self.sat_axes_maps)
            if self.scal_ensembles
            else [sat for sat in self.sat_axes_maps if sat in self.satfunc.columns]
        )

    @property
    def color_options(self):
        """Options to color by"""
        return ["ENSEMBLE", "CURVE", "SATNUM"]

    @property
    def ens_colors(self):
        return {
            ens: self.plotly_theme["layout"]["colorway"][self.ensembles.index(ens)]
            for ens in self.ensembles
        }

    @property
    def satnum_colors(self):
        return {
            satnum: self.plotly_theme["layout"]["colorway"][self.satnums.index(satnum)]
            for satnum in self.satnums
        }

    @property
    def tour_steps(self):
        return [
            {
                "id": self.uuid("layout"),
                "content": "Dashboard displaying relative permeability and capillary pressure"
                "data.",
            },
            {
                "id": self.uuid("graph"),
                "content": (
                    "Visualization of curves. "
                    "Different options can be set in the menu to the left."
                    "You can also toggle data on/off by clicking at the legend."
                ),
            },
            {
                "id": self.uuid("sataxis"),
                "content": (
                    "Choose saturation type for your x-axis. Will automatically change available "
                    "options in 'Curves'."
                ),
            },
            {
                "id": self.uuid("color_by"),
                "content": ("Choose basis for your colormap."),
            },
            {
                "id": self.uuid("ensemble"),
                "content": ("Select ensembles."),
            },
            {
                "id": self.uuid("curve_selector"),
                "content": (
                    "Choose curves. Capillary pressures and relative permeabilities will be shown"
                    " in separate plots."
                ),
            },
            {
                "id": self.uuid("satnum"),
                "content": ("Choose SATNUM regions."),
            },
            {
                "id": self.uuid("visualization"),
                "content": (
                    "Choose between different visualizations. 1. Show time series as "
                    "individual lines per realization. 2. Show statistical fanchart per ensemble."
                ),
            },
            {
                "id": self.uuid("linlog"),
                "content": ("Switch between linear and logarithmic y-axis."),
            },
            {
                "id": self.uuid("scal_selector"),
                "content": (
                    "Switch on/off SCAL reference data (requires that the optional scalfile is"
                    " defined)."
                ),
            },
        ]

    @staticmethod
    def set_grid_layout(columns, padding=0):
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
            "padding": f"{padding}px",
        }

    @property
    def layout(self):
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                wcc.Frame(
                    id=self.uuid("filters"),
                    style={"flex": "1", "height": "90vh"},
                    children=[
                        wcc.Selectors(
                            label="Selectors",
                            children=[
                                wcc.Dropdown(
                                    label="Saturation axis",
                                    id=self.uuid("sataxis"),
                                    clearable=False,
                                    options=[
                                        {
                                            "label": i.lower().capitalize(),
                                            "value": i,
                                        }
                                        for i in self.sat_axes
                                    ],
                                    value=self.sat_axes[0],
                                ),
                                wcc.Dropdown(
                                    label="Color by",
                                    id=self.uuid("color_by"),
                                    clearable=False,
                                    options=[
                                        {
                                            "label": i.lower().capitalize(),
                                            "value": i,
                                        }
                                        for i in self.color_options
                                    ],
                                    value=self.color_options[0],
                                ),
                                dcc.Store(
                                    id=self.uuid("stored_ensemble"),
                                    storage_type="session",
                                    data={},
                                ),
                                wcc.Dropdown(
                                    label="Ensembles",
                                    id=self.uuid("ensemble"),
                                    clearable=False,
                                    multi=True,
                                    options=[
                                        {"label": i, "value": i} for i in self.ensembles
                                    ],
                                    value=self.ensembles[0],
                                ),
                                wcc.Dropdown(
                                    label="Curves",
                                    id=self.uuid("curve"),
                                    clearable=False,
                                    multi=True,
                                ),
                                dcc.Store(
                                    id=self.uuid("stored_satnum"),
                                    storage_type="session",
                                    data={},
                                ),
                                wcc.Dropdown(
                                    label="Satnum",
                                    id=self.uuid("satnum"),
                                    clearable=False,
                                    multi=True,
                                    options=[
                                        {"label": i, "value": i} for i in self.satnums
                                    ],
                                    value=self.satnums[0],
                                ),
                            ],
                        ),
                        wcc.Selectors(
                            label="Visualization",
                            children=[
                                html.Div(
                                    wcc.RadioItems(
                                        label="Line traces",
                                        id=self.uuid("visualization"),
                                        className="block-options",
                                        options=[
                                            {
                                                "label": "Individual realizations",
                                                "value": "realizations",
                                            },
                                            {
                                                "label": "Statistical fanchart",
                                                "value": "statistics",
                                            },
                                        ],
                                        value="statistics",
                                    ),
                                    style={"display": "none"}
                                    if self.plugin_mode == "scal_scenarios"
                                    else {"display": "block"},
                                ),
                                html.Div(
                                    wcc.Checklist(
                                        label="SCAL recommendation",
                                        id=self.uuid("scal"),
                                        options=[
                                            {"label": val.capitalize(), "value": val}
                                            for val in ["pess", "base", "opt"]
                                        ],
                                        value=["pess", "base", "opt"]
                                        if self.scal_ensembles is not None
                                        else [],
                                    ),
                                    style={"display": "block"}
                                    if self.scal_ensembles is not None
                                    else {"display": "none"},
                                ),
                                wcc.RadioItems(
                                    label="Y-axis",
                                    id=self.uuid("linlog"),
                                    className="block-options",
                                    options=[
                                        {
                                            "label": "Linear",
                                            "value": "linear",
                                        },
                                        {
                                            "label": "Log (only kr)",
                                            "value": "log",
                                        },
                                    ],
                                    value="linear",
                                ),
                            ],
                        ),
                    ],
                ),
                wcc.Frame(
                    color="white",
                    highlight=False,
                    style={"flex": "4", "height": "90vh"},
                    children=wcc.Graph(style={"height": "88vh"}, id=self.uuid("graph")),
                ),
            ],
        )

    def set_callbacks(self, app):
        @app.callback(
            Output(self.uuid("graph"), "figure"),
            [
                Input(self.uuid("color_by"), "value"),
                Input(self.uuid("visualization"), "value"),
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("curve"), "value"),
                Input(self.uuid("satnum"), "value"),
                Input(self.uuid("sataxis"), "value"),
                Input(self.uuid("linlog"), "value"),
                Input(self.uuid("scal"), "value"),
            ],
        )
        def _update_graph(
            color_by,
            visualization,
            ensembles,
            curves,
            satnums,
            sataxis,
            linlog,
            selected_scal,
        ):
            if not curves or not satnums:  # Curve and satnum has to be defined
                raise PreventUpdate
            if ensembles is None:  # Allowing no ensembles to plot only SCAL data
                ensembles = []
            if not isinstance(ensembles, list):
                ensembles = [ensembles]
            if not isinstance(curves, list):
                curves = [curves]
            if not isinstance(satnums, list):
                satnums = [satnums]
            if color_by == "ENSEMBLE":
                colors = self.ens_colors
            elif color_by == "SATNUM":
                colors = self.satnum_colors
            else:
                colors = RelativePermeability.SCAL_COLORMAP
            nplots = (
                2
                if any(curve.startswith("PC") for curve in curves)
                and any(curve.startswith("KR") for curve in curves)
                else 1
            )
            layout = plot_layout(
                nplots, curves, sataxis, color_by, linlog, self.plotly_theme["layout"]
            )
            traces = []
            if self.plugin_mode == "distribution":
                df = filter_df(self.satfunc, ensembles, curves, satnums, sataxis)
                if visualization == "realizations" and not df.empty:
                    traces.extend(
                        add_realization_traces(
                            df, color_by, curves, sataxis, colors, nplots
                        )
                    )
                elif visualization == "statistics" and not df.empty:
                    traces.extend(
                        add_statistic_traces(
                            df, color_by, curves, sataxis, colors, nplots
                        )
                    )

            if selected_scal:
                if self.plugin_mode == "distribution":
                    scal_df = filter_scal_df(
                        self.scal_ensembles["SCAL"],
                        curves,
                        satnums,
                        sataxis,
                        selected_scal,
                    )
                    traces.extend(
                        add_scal_traces(
                            scal_df,
                            color_by,
                            curves,
                            sataxis,
                            colors,
                            nplots,
                            "SCAL",
                            self.plugin_mode,
                        )
                    )
                elif self.plugin_mode == "scal_scenarios":
                    for scal_scenario in ensembles:
                        scal_df = filter_scal_df(
                            self.scal_ensembles[scal_scenario],
                            curves,
                            satnums,
                            sataxis,
                            selected_scal,
                        )
                        traces.extend(
                            add_scal_traces(
                                scal_df,
                                color_by,
                                curves,
                                sataxis,
                                colors,
                                nplots,
                                scal_scenario,
                                self.plugin_mode,
                            )
                        )
                else:
                    raise ValueError(f"Invalid plugin_mode: {self.plugin_mode}")

            return {"data": traces, "layout": layout}

        @app.callback(
            [
                Output(self.uuid("ensemble"), "multi"),
                Output(self.uuid("ensemble"), "value"),
            ],
            [Input(self.uuid("color_by"), "value")],
            [State(self.uuid("stored_ensemble"), "data")],
        )
        def _set_ensemble_selector(color_by, stored_ensemble):
            """If ensemble is selected as color by, set the ensemble
            selector to allow multiple selections, else use stored_ensemble
            """

            if color_by == "ENSEMBLE":
                return True, self.ensembles

            return (
                False,
                stored_ensemble.get("ENSEMBLE", self.ensembles[0]),
            )

        @app.callback(
            [
                Output(self.uuid("satnum"), "multi"),
                Output(self.uuid("satnum"), "value"),
            ],
            [Input(self.uuid("color_by"), "value")],
            [State(self.uuid("stored_satnum"), "data")],
        )
        def _set_satnum_selector(color_by, stored_satnum):
            """If satnum is selected as color by, set the satnum
            selector to allow multiple selections, else use stored_satnum
            """

            if color_by == "SATNUM":
                return True, self.satnums

            return (
                False,
                stored_satnum.get("SATNUM", self.satnums[0]),
            )

        @app.callback(
            [
                Output(self.uuid("curve"), "value"),
                Output(self.uuid("curve"), "options"),
            ],
            [Input(self.uuid("sataxis"), "value")],
        )
        def _set_curve_selector(sataxis):
            """If satnum is selected as color by, set the satnum
            selector to allow multiple selections, else use stored_satnum
            """
            return (
                self.sat_axes_maps[sataxis],
                [
                    {
                        "label": i,
                        "value": i,
                    }
                    for i in self.sat_axes_maps[sataxis]
                ],
            )

    def add_webvizstore(self):
        if self.scal_scenarios:
            return [
                (
                    load_scal_recommendation,
                    [{"scalfile": Path(path)}],
                )
                for _, path in self.scal_scenarios.items()
            ]

        return [
            (
                load_satfunc,
                [
                    {
                        "ensemble_paths": self.ens_paths,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            )
            if self.relpermfile is None
            else (
                load_csv,
                [
                    {
                        "ensemble_paths": self.ens_paths,
                        "csv_file": self.relpermfile,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            )
        ] + (
            []
            if self.scalfile is None
            else [
                (
                    load_scal_recommendation,
                    [{"scalfile": self.scalfile, "sheet_name": self.sheet_name}],
                )
            ]
        )


def filter_df(df, ensembles, curves, satnums, sataxis):
    df = df.copy()
    df = df.loc[df["ENSEMBLE"].isin(ensembles)]
    df = df.loc[df["SATNUM"].isin(satnums)]
    columns = ["ENSEMBLE", "REAL", "SATNUM"] + [sataxis] + curves
    return df[columns].dropna()


def filter_scal_df(df, curves, satnums, sataxis, selected_scal):
    df = df.copy()
    df = df.loc[df["SATNUM"].isin(satnums)]
    df = df.loc[df["CASE"].isin(selected_scal)]
    columns = (
        ["SATNUM", "CASE"]
        + [sataxis]
        + [curve for curve in curves if curve in df.columns]
    )
    return df[columns].dropna()


def add_realization_traces(df, color_by, curves, sataxis, colors, nplots):
    """Renders line traces for individual realizations"""
    traces = []

    for curve_no, curve in enumerate(curves):
        yaxis = "y" if nplots == 1 or curve.startswith("KR") else "y2"
        xaxis = "x" if yaxis == "y" else "x2"
        if color_by == "CURVE":
            satnum = df["SATNUM"].iloc[0]
            ensemble = df["ENSEMBLE"].iloc[0]

            traces.extend(
                [
                    {
                        "type": "scatter",
                        "x": real_df[sataxis],
                        "y": real_df[curve],
                        "xaxis": xaxis,
                        "yaxis": yaxis,
                        "hovertext": (
                            f"{curve}, Satnum: {satnum}<br>"
                            f"Realization: {real}, Ensemble: {ensemble}"
                        ),
                        "name": curve,
                        "legendgroup": curve,
                        "marker": {
                            "color": colors.get(curve, colors[list(colors.keys())[0]])
                        },
                        "showlegend": real_no == 0,
                    }
                    for real_no, (real, real_df) in enumerate(df.groupby("REAL"))
                ]
            )
        else:
            constant_group = (
                df["SATNUM"].iloc[0]
                if color_by == "ENSEMBLE"
                else df["ENSEMBLE"].iloc[0]
            )
            traces.extend(
                [
                    {
                        "type": "scatter",
                        "x": real_df[sataxis],
                        "y": real_df[curve],
                        "xaxis": xaxis,
                        "yaxis": yaxis,
                        "hovertext": (
                            f"{curve}, Satnum: "
                            f"{group if color_by == 'SATNUM' else constant_group}<br>"
                            f"Realization: {real}, Ensemble: "
                            f"{group if color_by == 'ENSEMBLE' else constant_group}"
                        ),
                        "name": group,
                        "legendgroup": group,
                        "marker": {
                            "color": colors.get(group, colors[list(colors.keys())[-1]])
                        },
                        "showlegend": real_no == 0 and curve_no == 0,
                    }
                    for (group, grouped_df) in df.groupby(color_by)
                    for real_no, (real, real_df) in enumerate(
                        grouped_df.groupby("REAL")
                    )
                ]
            )
    return traces


# pylint: disable=too-many-locals
def add_scal_traces(
    df, color_by, curves, sataxis, colors, nplots, scenario_name, plugin_mode
):
    """Renders scal recommendation traces"""
    traces = []
    linestyle = {"pess": "dash", "base": "solid", "opt": "dot"}
    for curve_no, curve in enumerate(
        [curve for curve in curves if curve in df.columns]
    ):
        yaxis = "y" if nplots == 1 or curve.startswith("KR") else "y2"
        xaxis = "x" if yaxis == "y" else "x2"
        if plugin_mode == "scal_scenarios":
            if color_by == "CURVE":
                satnum = df["SATNUM"].iloc[0]
                traces.extend(
                    [
                        {
                            "type": "scatter",
                            "x": case_df[sataxis],
                            "y": case_df[curve],
                            "xaxis": xaxis,
                            "yaxis": yaxis,
                            "hovertext": (
                                f"{curve}, Satnum: {satnum}<br>"
                                f"{case.capitalize()}, scenario: {scenario_name}"
                            ),
                            "name": curve,
                            "legendgroup": curve,
                            "marker": {
                                "color": colors.get(
                                    curve, colors[list(colors.keys())[0]]
                                )
                            },
                            "line": {"dash": linestyle.get(case, "solid")},
                            "showlegend": case_no == 0,
                        }
                        for case_no, (case, case_df) in enumerate(df.groupby("CASE"))
                    ]
                )
            elif color_by == "SATNUM":
                traces.extend(
                    [
                        {
                            "type": "scatter",
                            "x": case_df[sataxis],
                            "y": case_df[curve],
                            "xaxis": xaxis,
                            "yaxis": yaxis,
                            "hovertext": (
                                f"{curve}, Satnum: {group}<br>"
                                f"{case.capitalize()}, Ensemble: {scenario_name}"
                            ),
                            "name": group,
                            "legendgroup": group,
                            "marker": {
                                "color": colors.get(
                                    group, colors[list(colors.keys())[-1]]
                                )
                            },
                            "line": {"dash": linestyle.get(case, "solid")},
                            "showlegend": case_no == 0 and curve_no == 0,
                        }
                        for (group, grouped_df) in df.groupby(color_by)
                        for case_no, (case, case_df) in enumerate(
                            grouped_df.groupby("CASE")
                        )
                    ]
                )
            else:
                constant_group = df["SATNUM"].iloc[0]
                traces.extend(
                    [
                        {
                            "type": "scatter",
                            "x": case_df[sataxis],
                            "y": case_df[curve],
                            "xaxis": xaxis,
                            "yaxis": yaxis,
                            "hovertext": (
                                f"{curve}, Satnum: {constant_group}<br>"
                                f"{case.capitalize()}, Ensemble: {scenario_name}"
                            ),
                            "name": scenario_name,
                            "legendgroup": scenario_name,
                            "marker": {
                                "color": colors.get(
                                    scenario_name, colors[list(colors.keys())[-1]]
                                )
                            },
                            "line": {"dash": linestyle.get(case, "solid")},
                            "showlegend": case_no == 0 and curve_no == 0,
                        }
                        for case_no, (case, case_df) in enumerate(df.groupby("CASE"))
                    ]
                )

        ##
        elif plugin_mode == "distribution":
            traces.extend(
                [
                    {
                        "type": "scatter",
                        "x": case_df[sataxis],
                        "y": case_df[curve],
                        "xaxis": xaxis,
                        "yaxis": yaxis,
                        "hovertext": (
                            f"{curve}, Satnum: {satnum}<br>"
                            f"{case.lower().capitalize()}"
                        ),
                        "name": scenario_name,
                        "legendgroup": scenario_name,
                        "marker": {
                            "color": "black",
                        },
                        "line": {"dash": linestyle.get(case, "solid")},
                        "showlegend": curve_no == 0 and satnum_no == 0 and case_no == 0,
                    }
                    for satnum_no, (satnum, satnum_df) in enumerate(
                        df.groupby("SATNUM")
                    )
                    for case_no, (case, case_df) in enumerate(satnum_df.groupby("CASE"))
                ]
            )
        else:
            raise ValueError(f"Invalid plugin_mode: {plugin_mode}")
    return traces


# pylint: disable=too-many-locals
def add_statistic_traces(df, color_by, curves, sataxis, colors, nplots):
    """Calculate statistics and call fanchart rendering"""
    # Switched P10 and P90 due to convetion in petroleum industry
    def p10(x):
        return np.nanpercentile(x, q=90)

    def p90(x):
        return np.nanpercentile(x, q=10)

    traces = []
    for ens_no, (ens, ens_df) in enumerate(
        df[["ENSEMBLE", "REAL", "SATNUM", sataxis] + curves].groupby(["ENSEMBLE"])
    ):
        for satnum_no, (satnum, satnum_df) in enumerate(
            ens_df[["REAL", "SATNUM", sataxis] + curves].groupby("SATNUM")
        ):
            satnum_df_shared_axis = satnum_df[
                satnum_df.groupby(sataxis)["REAL"].transform("nunique")
                == satnum_df["REAL"].nunique()
            ]
            df_stat = (
                satnum_df_shared_axis.groupby(sataxis)
                .agg([np.nanmean, np.nanmin, np.nanmax, p10, p90])
                .stack()
                .swaplevel()
            )
            for curve_no, curve in enumerate(curves):
                yaxis = "y" if nplots == 1 or curve.startswith("KR") else "y2"
                xaxis = "x" if yaxis == "y" else "x2"
                legend_group = (
                    curve
                    if color_by == "CURVE"
                    else ens
                    if color_by == "ENSEMBLE"
                    else satnum
                )
                show_legend = (
                    bool(color_by == "CURVE" and ens_no == 0 and satnum_no == 0)
                    or bool(color_by == "ENSEMBLE" and curve_no == 0 and satnum_no == 0)
                    or bool(color_by == "SATNUM" and curve_no == 0 and ens_no == 0)
                )
                traces.extend(
                    _get_fanchart_traces(
                        df_stat[curve],
                        colors.get(legend_group, colors[list(colors.keys())[0]]),
                        xaxis,
                        yaxis,
                        legend_group,
                        show_legend,
                        curve,
                        ens,
                        satnum,
                    )
                )
    return traces


# pylint: disable=too-many-arguments
def _get_fanchart_traces(
    curve_stats,
    color,
    xaxis,
    yaxis,
    legend_group: str,
    show_legend: bool,
    curve,
    ens,
    satnum,
):
    """Renders a fanchart"""

    # Retrieve indices from one of the keys in series
    x = curve_stats["nanmax"].index.tolist()
    data = FanchartData(
        samples=x,
        low_high=LowHighData(
            low_data=curve_stats["p90"].values,
            low_name="P90",
            high_data=curve_stats["p10"].values,
            high_name="P10",
        ),
        minimum_maximum=MinMaxData(
            minimum=curve_stats["nanmin"].values,
            maximum=curve_stats["nanmax"].values,
        ),
        free_line=FreeLineData("Mean", curve_stats["nanmean"].values),
    )

    hovertemplate = f"{curve} <br>" f"Ensemble: {ens}, Satnum: {satnum}"

    return get_fanchart_traces(
        data=data,
        hex_color=color,
        legend_group=legend_group,
        xaxis=xaxis,
        yaxis=yaxis,
        hovertext=hovertemplate,
        show_legend=show_legend,
    )


@CACHE.memoize()
def plot_layout(nplots, curves, sataxis, color_by, linlog, theme):
    """
    Constructing plot layout from scratch as it is more responsive than plotly subplots package.
    """
    titles = (
        ["Relative Permeability", "Capillary Pressure"]
        if nplots == 2
        else ["Relative Permeability"]
        if any(curve.startswith("KR") for curve in curves)
        else ["Capillary Pressure"]
    )
    layout = theme.copy()
    layout.update(
        {
            "hovermode": "closest",
            "uirevision": f"sa:{sataxis}_{linlog}_curves:{'_'.join(sorted(curves))}",
        }
    )
    # create subplots
    layout.update(
        {
            "annotations": [
                {
                    "showarrow": False,
                    "text": titles[0],
                    "x": 0.5,
                    "xanchor": "center",
                    "xref": "paper",
                    "y": 1.0,
                    "yanchor": "bottom",
                    "yref": "paper",
                    "font": {"size": 16},
                }
            ],
        }
        if nplots == 1
        else {
            "annotations": [
                {
                    "showarrow": False,
                    "text": titles[0],
                    "x": 0.5,
                    "xanchor": "center",
                    "xref": "paper",
                    "y": 1.0,
                    "yanchor": "bottom",
                    "yref": "paper",
                    "font": {"size": 16},
                },
                {
                    "showarrow": False,
                    "text": titles[1],
                    "x": 0.5,
                    "xanchor": "center",
                    "xref": "paper",
                    "y": 0.475,
                    "yanchor": "bottom",
                    "yref": "paper",
                    "font": {"size": 16},
                },
            ],
        }
        if nplots == 2
        else {}
    )

    layout["legend"] = {"title": {"text": color_by.lower().capitalize()}}
    # format axes
    if nplots == 1:
        layout.update(
            {
                "xaxis": {
                    "automargin": True,
                    "ticks": "",
                    "zeroline": False,
                    "range": [0, 1],
                    "anchor": "y",
                    "domain": [0.0, 1.0],
                    "title": {"text": sataxis.lower().capitalize(), "standoff": 15},
                    "showgrid": False,
                    "tickmode": "auto",
                    "exponentformat": "e",
                },
                "yaxis": {
                    "automargin": True,
                    "ticks": "",
                    "zeroline": False,
                    "anchor": "x",
                    "domain": [0.0, 1.0],
                    "showgrid": False,
                    "exponentformat": "e",
                },
                "margin": {"t": 20, "b": 0},
            }
        )
        if any(curve.startswith("KR") for curve in curves):
            layout["yaxis"].update({"title": {"text": "kr"}, "type": linlog})
        else:
            layout["yaxis"].update({"title": {"text": "Pc"}, "type": "linear"})

    elif nplots == 2:
        layout.update(
            {
                "xaxis": {
                    "automargin": True,
                    "zeroline": False,
                    "anchor": "y",
                    "domain": [0.0, 1.0],
                    "matches": "x2",
                    "showticklabels": False,
                    "showgrid": False,
                    "exponentformat": "e",
                },
                "xaxis2": {
                    "automargin": True,
                    "ticks": "",
                    "showticklabels": True,
                    "zeroline": False,
                    "range": [0, 1],
                    "anchor": "y2",
                    "domain": [0.0, 1.0],
                    "title": {"text": sataxis.lower().capitalize()},
                    "showgrid": False,
                    "exponentformat": "e",
                },
                "yaxis": {
                    "automargin": True,
                    "ticks": "",
                    "zeroline": False,
                    "anchor": "x",
                    "domain": [0.525, 1.0],
                    "title": {"text": "kr"},
                    "type": linlog,
                    "showgrid": False,
                    "exponentformat": "e",
                },
                "yaxis2": {
                    "automargin": True,
                    "ticks": "",
                    "zeroline": False,
                    "anchor": "x2",
                    "domain": [0.0, 0.475],
                    "title": {"text": "Pc"},
                    "type": "linear",
                    "showgrid": False,
                    "exponentformat": "e",
                },
                "margin": {"t": 20, "b": 0},
            }
        )
    return layout



# pylint: disable=too-many-lines
import copy
import datetime
import json
import sys
import warnings
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Union

import dash
import dash_bootstrap_components as dbc
import numpy as np
import pandas as pd
import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import Dash, Input, Output, State, dcc, html
from dash.exceptions import PreventUpdate
from plotly.subplots import make_subplots
from webviz_config import EncodedFile, WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.deprecation_decorators import deprecated_plugin
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from webviz_config.webviz_store import webvizstore
from webviz_subsurface_components import ExpressionInfo, ExternalParseData

import webviz_subsurface
from webviz_subsurface._models import (
    EnsembleSetModel,
    caching_ensemble_set_model_factory,
)

from .._abbreviations.reservoir_simulation import (
    historical_vector,
    simulation_unit_reformat,
    simulation_vector_description,
)
from .._datainput.from_timeseries_cumulatives import (
    calc_from_cumulatives,
    rename_vec_from_cum,
)
from .._utils.fanchart_plotting import (
    FanchartData,
    FreeLineData,
    LowHighData,
    MinMaxData,
    get_fanchart_traces,
)
from .._utils.simulation_timeseries import (
    add_statistics_traces,
    calc_series_statistics,
    check_and_format_observations,
    date_to_interval_conversion,
    get_simulation_line_shape,
    render_hovertemplate,
    set_simulation_line_shape_fallback,
)
from .._utils.unique_theming import unique_colors
from .._utils.vector_calculator import (
    expressions_from_config,
    get_calculated_units,
    get_calculated_vector_df,
    get_expression_from_name,
    get_selected_expressions,
    get_vector_definitions_from_expressions,
    validate_predefined_expression,
)
from .._utils.vector_selector import (
    add_vector_to_vector_selector_data,
    is_vector_name_in_vector_selector_data,
)


def _check_plugin_options(options: Optional[dict]) -> Optional[Tuple[str, str]]:
    if options:
        if "vector1" in options or "vector2" in options or "vector3" in options:
            return (
                "Please use 'vectors' instead of 'vector1/2/3' in this plugin's options.",
                "Single vector options ('vector1', 'vector2', 'vector3')"
                " have been replaced with a vectors list.",
            )
    return None


# pylint: disable = too-many-instance-attributes
@deprecated_plugin(
    "This plugin has been replaced by the faster, "
    "more flexible and less memory hungry plugin `SimulationTimeSeries`"
)
class ReservoirSimulationTimeSeries(WebvizPluginABC):
    """Visualizes reservoir simulation time series data for FMU ensembles.

**Features**
* Visualization of realization time series as line charts.
* Visualization of ensemble time series statistics as line or fan charts.
* Visualization of single date ensemble statistics as histograms.
* Calculation and visualization of delta ensembles.
* Calculation and visualization of average rates and cumulatives over a specified time interval.
* Download of visualized data to csv files (except histogram data).

---
**Two main options for input data: Aggregated and read from UNSMRY.**

**Using aggregated data**
* **`csvfile`:** Aggregated csv file with `REAL`, `ENSEMBLE`, \
    `DATE` and vector columns.

**Using simulation time series data directly from `UNSMRY` files**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.
* **`column_keys`:** List of vectors to extract. If not given, all vectors \
    from the simulations will be extracted. Wild card asterisk `*` can be used.
* **`sampling`:** Time separation between extracted values. Can be e.g. `monthly` (default) or \
    `yearly`.

**Common optional settings for both input options**
* **`obsfile`**: File with observations to plot together with the relevant time series. \
(absolute path or relative to config file).
* **`options`:** Options to initialize plots with:
    * `vector1` : First vector to display
    * `vector2` : Second vector to display
    * `vector3` : Third vector to display
    * `visualization` : `realizations`, `statistics` or `fanchart`
    * `date` : Date to show by default in histograms
* **`line_shape_fallback`:** Fallback interpolation method between points. Vectors identified as \
    rates or phase ratios are always backfilled, vectors identified as cumulative (totals) are \
    always linearly interpolated. The rest use the fallback.
    Supported options:
    * `linear` (default)
    * `backfilled`
    * `hv`, `vh`, `hvh`, `vhv` and `spline` (regular Plotly options).

**Calculated vector expressions**
* **`predefined_expressions`:** yaml file with pre-defined expressions

---

?> Vectors that are identified as historical vectors (e.g. FOPTH is the history of FOPT) will \
be plotted together with their non-historical counterparts as reference lines, and they are \
therefore not selectable as vectors to plot initially.

?> The `obsfile` is a common (optional) file for all ensembles, which can be \
converted from e.g. ERT and ResInsight formats using the [fmuobs]\
(https://equinor.github.io/subscript/scripts/fmuobs.html) script. \
[An example of the format can be found here]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/reek_history_match/share/\
observations/observations.yml).

!> It is **strongly recommended** to keep the data frequency to a regular frequency (like \
`monthly` or `yearly`). This applies to both csv input and when reading from `UNSMRY` \
(controlled by the `sampling` key). This is because the statistics and fancharts are calculated \
per DATE over all realizations in an ensemble, and the available dates should therefore not \
differ between individual realizations of an ensemble.

**Using aggregated data**

The `csvfile` must have columns `ENSEMBLE`, `REAL` and `DATE` in addition to the individual
vectors.
* [Example of aggregated file]\
(https://github.com/equinor/webviz-subsurface-testdata/blob/master/aggregated_data/smry.csv).

**Using simulation time series data directly from `.UNSMRY` files**

Vectors are extracted automatically from the `UNSMRY` files in the individual realizations,
using the `fmu-ensemble` library.

?> Using the `UNSMRY` method will also extract metadata like units, and whether the vector is a \
rate, a cumulative, or historical. Units are e.g. added to the plot titles, while rates and \
cumulatives are used to decide the line shapes in the plot. Aggregated data may on the other \
speed up the build of the app, as processing of `UNSMRY` files can be slow for large models. \
Using this method is required to use the average rate and interval cumulative functionalities, \
as they require identification of vectors that are cumulatives.

!> The `UNSMRY` files are auto-detected by `fmu-ensemble` in the `eclipse/model` folder of the \
individual realizations. You should therefore not have more than one `UNSMRY` file in this \
folder, to avoid risk of not extracting the right data.
"""

    ENSEMBLE_COLUMNS = ["REAL", "ENSEMBLE", "DATE"]

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-statements
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile: Path = None,
        ensembles: list = None,
        obsfile: Path = None,
        column_keys: list = None,
        sampling: str = "monthly",
        options: dict = None,
        predefined_expressions: str = None,
        line_shape_fallback: str = "linear",
    ):

        super().__init__()

        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "css"
            / "block_options.css"
        )

        # Temporary css, pending on new wcc modal component.
        # See: https://github.com/equinor/webviz-core-components/issues/163
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent / "_assets" / "css" / "modal.css"
        )

        self.csvfile = csvfile
        self.obsfile = obsfile
        self.time_index = sampling
        self.column_keys = column_keys
        if csvfile and ensembles:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile" or "ensembles"'
            )
        self.observations = {}
        if obsfile:
            self.observations = check_and_format_observations(get_path(self.obsfile))

        self.smry: pd.DataFrame
        self.smry_meta: Union[pd.DataFrame, None]

        if csvfile:
            self.smry = read_csv(csvfile)
            self.smry_meta = None
            # Check of time_index for data to use in resampling. Quite naive as it only checks for
            # unique values of the DATE column, and not per realization.
            #
            # Currently not necessary as we don't allow resampling for average rates and intervals
            # unless we have metadata, which csvfile input currently doesn't support.
            # See: https://github.com/equinor/webviz-subsurface/issues/402
            self.time_index = pd.infer_freq(
                sorted(pd.to_datetime(self.smry["DATE"]).unique())
            )
        elif ensembles:
            self.emodel: EnsembleSetModel = (
                caching_ensemble_set_model_factory.get_or_create_model(
                    ensemble_paths={
                        ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                        for ens in ensembles
                    },
                    time_index=self.time_index,
                    column_keys=self.column_keys,
                )
            )
            self.smry = self.emodel.get_or_load_smry_cached()
            self.smry_meta = self.emodel.load_smry_meta()
        else:
            raise ValueError(
                'Incorrent arguments. Either provide a "csvfile" or "ensembles"'
            )
        if any(col.startswith(("AVG_", "INTVL_")) for col in self.smry.columns):
            raise ValueError(
                "Your data set includes time series vectors which have names starting with"
                "'AVG_' and/or 'INTVL_'. These prefixes are not allowed, as they are used"
                "internally in the plugin."
            )
        self.smry_cols: List[str] = [
            c
            for c in self.smry.columns
            if c not in ReservoirSimulationTimeSeries.ENSEMBLE_COLUMNS
            and historical_vector(c, self.smry_meta, False) not in self.smry.columns
        ]

        self.vector_data: list = []
        for vec in self.smry_cols:
            split = vec.split(":")
            add_vector_to_vector_selector_data(
                self.vector_data, vec, simulation_vector_description(split[0])
            )

            if (
                self.smry_meta is not None
                and self.smry_meta.is_total[vec]
                and self.time_index is not None
            ):
                # Get the likely name for equivalent rate vector and make dropdown options.
                # Requires that the time_index was either defined or possible to infer.
                avgrate_vec = rename_vec_from_cum(vector=vec, as_rate=True)
                interval_vec = rename_vec_from_cum(vector=vec, as_rate=False)

                avgrate_split = avgrate_vec.split(":")
                interval_split = interval_vec.split(":")

                add_vector_to_vector_selector_data(
                    self.vector_data,
                    avgrate_vec,
                    f"{simulation_vector_description(avgrate_split[0])} ({avgrate_vec})",
                )
                add_vector_to_vector_selector_data(
                    self.vector_data,
                    interval_vec,
                    f"{simulation_vector_description(interval_split[0])} ({interval_vec})",
                )

        self.ensembles = list(self.smry["ENSEMBLE"].unique())
        self.theme = webviz_settings.theme

        self.plot_options = options if options else {}
        if "vectors" not in self.plot_options:
            self.plot_options["vectors"] = []
        for vector in [
            vector
            for vector in ["vector1", "vector2", "vector3"]
            if vector in self.plot_options
        ]:
            self.plot_options["vectors"].append(self.plot_options[vector])
        self.plot_options["vectors"] = self.plot_options["vectors"][:3]

        self.plot_options["date"] = (
            str(self.plot_options.get("date"))
            if self.plot_options.get("date")
            else str(max(self.smry["DATE"]))
        )
        self.line_shape_fallback = set_simulation_line_shape_fallback(
            line_shape_fallback
        )

        # Retreive predefined expressions from configuration and validate
        self.predefined_expressions_path = (
            None
            if predefined_expressions is None
            else webviz_settings.shared_settings["predefined_expressions"][
                predefined_expressions
            ]
        )
        self.predefined_expressions = expressions_from_config(
            get_path(self.predefined_expressions_path)
            if self.predefined_expressions_path
            else None
        )
        for expression in self.predefined_expressions:
            valid, message = validate_predefined_expression(
                expression, self.vector_data
            )
            if not valid:
                warnings.warn(message)
            expression["isValid"] = valid

        # Create initial vector selector data
        self.initial_vector_selector_data = copy.deepcopy(self.vector_data)
        self._add_expressions_to_vector_data(
            self.initial_vector_selector_data, self.predefined_expressions
        )

        # Check if initially plotted vectors exist in data, raise ValueError if not.
        missing_vectors = (
            [
                value
                for value in self.plot_options["vectors"]
                if value not in self.smry_cols
            ]
            if "vectors" in self.plot_options
            else []
        )
        if missing_vectors:
            raise ValueError(
                f"Cannot find: {', '.join(missing_vectors)} to plot initially in "
                "ReservoirSimulationTimeSeries. Check that the vectors exist in your data, and "
                "that they are not missing in a non-default column_keys list in the yaml config "
                "file."
            )
        self.allow_delta = len(self.ensembles) > 1
        self.set_callbacks(app)

    @staticmethod
    def _add_expression(
        expression_data: list,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        description_str = description if description is not None else ""
        add_vector_to_vector_selector_data(
            vector_selector_data=expression_data,
            vector=name,
            description=description_str,
            description_at_last_node=True,
        )

    @property
    def ens_colors(self) -> dict:
        return unique_colors(self.ensembles, self.theme)

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.uuid("layout"),
                "content": "Dashboard displaying reservoir simulation time series.",
            },
            {
                "id": self.uuid("graph"),
                "content": (
                    "Visualization of selected time series. "
                    "Different options can be set in the menu to the left."
                ),
            },
            {
                "id": self.uuid("ensemble"),
                "content": (
                    "Display time series from one or several ensembles. "
                    "Different ensembles will be overlain in the same plot."
                ),
            },
            {
                "id": self.uuid("vectors"),
                "content": (
                    "Display up to three different time series. "
                    "Each time series will be visualized in a separate plot. "
                    "Vectors prefixed with AVG_ and INTVL_ are calculated in the fly "
                    "from cumulative vectors, providing average rates and interval cumulatives "
                    "over a time interval that can be defined in the menu."
                ),
            },
            {
                "id": self.uuid("visualization"),
                "content": (
                    "Choose between different visualizations. 1. Show time series as "
                    "individual lines per realization. 2. Show statistical lines per "
                    "ensemble. 3. Show statistical fanchart per ensemble"
                    "per date. Select a date by clicking in the plot."
                ),
            },
            {
                "id": self.uuid("options"),
                "content": (
                    "Various plot options: Whether to include history trace, add histogram, "
                    "and which statistics to show if statistical lines is chosen as visualization."
                ),
            },
            {
                "id": self.uuid("cum_interval"),
                "content": (
                    "Defines the time interval the average rates (prefixed AVG_) and interval "
                    "cumulatives (prefixed INTVL_) are calculated over. Disabled unless at least "
                    "one time series dependent on the interval setting is chosen."
                    "The option might be completely hidden if the data input does not support "
                    "calculation from cumulatives."
                ),
            },
            {
                "id": self.uuid("vector_calculator_selector"),
                "content": (
                    "Create mathematical expressions with provided vector time series. "
                    "Parsing of the mathematical expression is handled and will give feedback "
                    "when entering invalid expressions. "
                    "The expressions are calculated on the fly and can be selected among the time "
                    "series to be shown in the plots."
                ),
            },
        ]

    @staticmethod
    def set_grid_layout(columns: str, padding: int = 0) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
            "padding": f"{padding}px",
        }

    @property
    def time_interval_options(self) -> List[str]:
        if self.time_index == "daily":
            return ["daily", "monthly", "yearly"]
        if self.time_index == "monthly":
            return ["monthly", "yearly"]
        if self.time_index == "yearly":
            return ["yearly"]
        return []

    @property
    def delta_layout(self) -> html.Div:
        show_delta = "block" if self.allow_delta else "none"
        return html.Div(
            children=[
                html.Div(
                    style={"display": show_delta},
                    children=wcc.RadioItems(
                        label="Mode",
                        id=self.uuid("mode"),
                        style={"marginBottom": "0.5vh"},
                        options=[
                            {
                                "label": "Individual ensembles",
                                "value": "ensembles",
                            },
                            {
                                "label": "Delta between ensembles",
                                "value": "delta_ensembles",
                            },
                        ],
                        value="ensembles",
                    ),
                ),
                wcc.Dropdown(
                    wrapper_id=self.uuid("show_ensembles"),
                    label="Selected ensembles",
                    id=self.uuid("ensemble"),
                    clearable=False,
                    multi=True,
                    options=[{"label": i, "value": i} for i in self.ensembles],
                    value=[self.ensembles[0]],
                ),
                html.Div(
                    id=self.uuid("calc_delta"),
                    style={"display": "none"},
                    children=[
                        wcc.Label("Selected ensemble delta (A-B):"),
                        wcc.FlexBox(
                            children=[
                                wcc.FlexColumn(
                                    min_width="100px",
                                    children=wcc.Dropdown(
                                        label="Ensemble A",
                                        id=self.uuid("base_ens"),
                                        clearable=False,
                                        options=[
                                            {"label": i, "value": i}
                                            for i in self.ensembles
                                        ],
                                        value=self.ensembles[0],
                                    ),
                                ),
                                wcc.FlexColumn(
                                    min_width="100px",
                                    children=wcc.Dropdown(
                                        label="Ensemble B",
                                        id=self.uuid("delta_ens"),
                                        clearable=False,
                                        options=[
                                            {"label": i, "value": i}
                                            for i in self.ensembles
                                        ],
                                        value=self.ensembles[-1],
                                    ),
                                ),
                            ],
                        ),
                    ],
                ),
            ],
        )

    @property
    def from_cumulatives_layout(self) -> html.Div:
        return html.Div(
            style=(
                {}
                if len(self.time_interval_options) > 0 and self.smry_meta is not None
                else {"display": "none"}
            ),
            children=[
                wcc.Label("Calculated from cumulatives:"),
                wcc.Label(
                    "Average (AVG_) and interval (INTVL_) time series",
                    style={"font-style": "italic"},
                ),
                html.Div(
                    wcc.RadioItems(
                        id=self.uuid("cum_interval"),
                        className="block-options",
                        options=[
                            {
                                "label": (f"{i.lower().capitalize()}"),
                                "value": i,
                                "disabled": False,
                            }
                            for i in self.time_interval_options
                        ],
                        value=self.time_index,
                    ),
                ),
            ],
        )

    @property
    def modal_vector_calculator_layout(self) -> html.Div:
        return dbc.Modal(
            style={"marginTop": "20vh", "width": "1300px"},
            children=[
                dbc.ModalHeader("Vector Calculator"),
                dbc.ModalBody(
                    html.Div(
                        id=self.uuid("vector_calculator_modal_body"),
                        children=[
                            wsc.VectorCalculator(
                                id=self.uuid("vector_calculator"),
                                vectors=self.vector_data,
                                expressions=self.predefined_expressions,
                            )
                        ],
                    ),
                ),
            ],
            id=self.uuid("modal_vector_calculator"),
            size="lg",
            centered=True,
        )

    @property
    def layout(self) -> wcc.FlexBox:
        return wcc.FlexBox(
            id=self.uuid("layout"),
            children=[
                wcc.FlexColumn(
                    children=wcc.Frame(
                        style={"height": "90vh"},
                        children=[
                            wcc.Selectors(
                                label="Ensembles", children=[self.delta_layout]
                            ),
                            wcc.Selectors(
                                label="Time series",
                                children=wsc.VectorSelector(
                                    id=self.uuid("vectors"),
                                    maxNumSelectedNodes=3,
                                    data=self.initial_vector_selector_data,
                                    placeholder="Add new vector...",
                                    persistence=True,
                                    persistence_type="session",
                                    selectedTags=self.plot_options.get(
                                        "vectors", [self.smry_cols[0]]
                                    ),
                                    numSecondsUntilSuggestionsAreShown=0.5,
                                    lineBreakAfterTag=True,
                                    customVectorDefinitions=(
                                        get_vector_definitions_from_expressions(
                                            self.predefined_expressions
                                        )
                                    ),
                                ),
                            ),
                            wcc.Selectors(
                                label="Visualization",
                                id=self.uuid("visualization"),
                                children=[
                                    wcc.RadioItems(
                                        id=self.uuid("statistics"),
                                        options=[
                                            {
                                                "label": "Individual realizations",
                                                "value": "realizations",
                                            },
                                            {
                                                "label": "Statistical lines",
                                                "value": "statistics",
                                            },
                                            {
                                                "label": "Statistical fanchart",
                                                "value": "fanchart",
                                            },
                                        ],
                                        value=self.plot_options.get(
                                            "visualization", "statistics"
                                        ),
                                    ),
                                ],
                            ),
                            wcc.Selectors(
                                label="Options",
                                id=self.uuid("options"),
                                children=[
                                    wcc.Checklist(
                                        id=self.uuid("trace_options"),
                                        options=[
                                            {"label": val, "value": val}
                                            for val in ["History", "Histogram"]
                                        ],
                                        value=["History"],
                                    ),
                                    html.Div(
                                        id=self.uuid("view_stat_options"),
                                        style={"display": "block"}
                                        if "statistics"
                                        in self.plot_options.get("visualization", "")
                                        else {"display": "none"},
                                        children=[
                                            wcc.Checklist(
                                                id=self.uuid("stat_options"),
                                                options=[
                                                    {"label": val, "value": val}
                                                    for val in [
                                                        "Mean",
                                                        "P10 (high)",
                                                        "P50 (median)",
                                                        "P90 (low)",
                                                        "Maximum",
                                                        "Minimum",
                                                    ]
                                                ],
                                                value=[
                                                    "Mean",
                                                    "P10 (high)",
                                                    "P90 (low)",
                                                ],
                                            ),
                                        ],
                                    ),
                                ],
                            ),
                            wcc.Selectors(
                                label="Calculations",
                                children=[self.from_cumulatives_layout],
                            ),
                            wcc.Selectors(
                                label="Vector Calculator",
                                id=self.uuid("vector_calculator_selector"),
                                open=False,
                                children=[
                                    dbc.Button(
                                        "Vector Calculator",
                                        id=self.uuid("vector_calculator_open_btn"),
                                    ),
                                ],
                            ),
                        ],
                    )
                ),
                self.modal_vector_calculator_layout,
                wcc.FlexColumn(
                    flex=4,
                    children=[
                        wcc.Frame(
                            style={"height": "90vh"},
                            highlight=False,
                            color="white",
                            children=wcc.Graph(
                                style={"height": "85vh"},
                                id=self.uuid("graph"),
                            ),
                        ),
                        dcc.Store(
                            id=self.uuid("date"),
                            storage_type="session",
                            data=json.dumps(self.plot_options.get("date", None)),
                        ),
                    ],
                ),
                dcc.Store(
                    id=self.uuid("vector_calculator_expressions_modal_open"),
                    data=self.predefined_expressions,
                ),
                dcc.Store(
                    id=self.uuid("vector_calculator_expressions"),
                    data=self.predefined_expressions,
                ),
            ],
        )

    @staticmethod
    def _get_valid_vector_selections(
        vector_data: list,
        selected_vectors: List[str],
        new_expressions: List[ExpressionInfo],
        existing_expressions: List[ExpressionInfo],
    ) -> List[str]:
        valid_selections: List[str] = []
        for vector in selected_vectors:
            new_vector: Optional[str] = vector

            # Get id if vector is among existing expressions
            dropdown_id = next(
                (elm["id"] for elm in existing_expressions if elm["name"] == vector),
                None,
            )
            # Find id among new expressions to get new/edited name
            if dropdown_id:
                new_vector = next(
                    (
                        elm["name"]
                        for elm in new_expressions
                        if elm["id"] == dropdown_id
                    ),
                    None,
                )

            # Append if vector name exist among data
            if new_vector is not None and is_vector_name_in_vector_selector_data(
                new_vector, vector_data
            ):
                valid_selections.append(new_vector)

        return valid_selections

    def _add_expressions_to_vector_data(
        self, vector_data: list, expressions: List[ExpressionInfo]
    ) -> None:
        for expression in expressions:
            if not expression["isValid"]:
                continue

            name = expression["name"]
            description = None
            if "description" in expression.keys():
                description = expression["description"]

            self._add_expression(vector_data, name, description)

    # pylint: disable=too-many-statements
    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.uuid("graph"), "figure"),
            [
                Input(self.uuid("vectors"), "selectedNodes"),
                Input(self.uuid("ensemble"), "value"),
                Input(self.uuid("mode"), "value"),
                Input(self.uuid("base_ens"), "value"),
                Input(self.uuid("delta_ens"), "value"),
                Input(self.uuid("statistics"), "value"),
                Input(self.uuid("cum_interval"), "value"),
                Input(self.uuid("date"), "data"),
                Input(self.uuid("trace_options"), "value"),
                Input(self.uuid("stat_options"), "value"),
                Input(
                    self.uuid("vector_calculator_expressions"),
                    "data",
                ),
            ],
        )
        # pylint: disable=too-many-arguments, too-many-locals, too-many-branches
        def _update_graph(
            vectors: List[str],
            ensembles: List[str],
            calc_mode: str,
            base_ens: str,
            delta_ens: str,
            visualization: str,
            cum_interval: str,
            stored_date: str,
            trace_options: List[str],
            stat_options: List[str],
            expressions: List[ExpressionInfo],
        ) -> dict:
            """Callback to update all graphs based on selections"""

            if not isinstance(ensembles, list):
                raise TypeError("ensembles should always be of type list")

            if calc_mode not in ["ensembles", "delta_ensembles"]:
                raise PreventUpdate

            if vectors is None:
                vectors = self.plot_options.get("vectors", [self.smry_cols[0]])

            # Retreive selected expressions
            selected_expressions = get_selected_expressions(expressions, vectors)
            calculated_units = pd.Series()
            if self.smry_meta is not None:
                calculated_units = get_calculated_units(
                    selected_expressions, self.smry_meta["unit"]
                )

            # Synthesize ensembles list for delta mode
            if calc_mode == "delta_ensembles":
                ensembles = [base_ens, delta_ens]

            # Retrieve previous/current selected date
            date = json.loads(stored_date) if stored_date else None

            # Titles for subplots
            # TODO(Sigurd)
            # Added None to union since date_to_interval_conversion() may return None.
            # Need input on what should be done since a None title is probably not what we want
            titles: List[Union[str, None]] = []
            for vec in vectors:
                if sys.version_info >= (3, 9):
                    unit_vec = vec.removeprefix("AVG_").removeprefix("INTVL_")
                else:
                    unit_vec = (
                        vec[4:]
                        if vec.startswith("AVG_")
                        else (vec[6:] if vec.startswith("INTVL_") else vec)
                    )
                if self.smry_meta is None:
                    titles.append(simulation_vector_description(vec))
                elif vec in calculated_units:
                    titles.append(f"{vec}" f" [{calculated_units[vec]}]")
                else:
                    titles.append(
                        f"{simulation_vector_description(vec)}"
                        f" [{simulation_unit_reformat(self.smry_meta.unit[unit_vec])}]"
                    )
                if "Histogram" in trace_options:
                    titles.append(
                        date_to_interval_conversion(
                            date=date,
                            vector=vec,
                            interval=cum_interval,
                            as_date=False,
                        )
                    )

            # Make a plotly subplot figure
            fig = make_subplots(
                rows=max(1, len(vectors)),
                cols=2 if "Histogram" in trace_options else 1,
                shared_xaxes=True,
                vertical_spacing=0.05,
                subplot_titles=titles if titles else ["No vector selected"],
            )

            # Loop through each vector and calculate relevant plot
            legends = []
            dfs = calculate_vector_dataframes(
                smry=self.smry,
                smry_meta=self.smry_meta,
                ensembles=ensembles,
                vectors=vectors,
                selected_expressions=selected_expressions,
                calc_mode=calc_mode,
                visualization=visualization,
                time_index=self.time_index,
                cum_interval=cum_interval,
            )
            for i, vector in enumerate(vectors):
                if dfs[vector]["data"].empty:
                    continue
                line_shape = get_simulation_line_shape(
                    line_shape_fallback=self.line_shape_fallback,
                    vector=vector,
                    smry_meta=self.smry_meta,
                )
                if visualization == "fanchart":
                    traces = _get_fanchart_traces(
                        dfs[vector]["stat"],
                        vector,
                        colors=self.ens_colors,
                        line_shape=line_shape,
                        interval=cum_interval,
                    )
                elif visualization == "statistics":
                    traces = _add_statistics_traces(
                        dfs[vector]["stat"],
                        vector,
                        colors=self.ens_colors,
                        line_shape=line_shape,
                        interval=cum_interval,
                        stat_options=stat_options,
                    )

                elif visualization == "realizations":
                    traces = add_realization_traces(
                        dfs[vector]["data"],
                        vector,
                        colors=self.ens_colors,
                        line_shape=line_shape,
                        interval=cum_interval,
                    )
                else:
                    raise PreventUpdate

                if "Histogram" in trace_options:
                    histdata = add_histogram_traces(
                        dfs[vector]["data"],
                        vector,
                        date=date,
                        colors=self.ens_colors,
                        interval=cum_interval,
                    )
                    for trace in histdata:
                        fig.add_trace(trace, i + 1, 2)
                if "History" in trace_options:
                    historical_vector_name = historical_vector(
                        vector=vector, smry_meta=self.smry_meta
                    )
                    if (
                        historical_vector_name
                        and historical_vector_name in dfs[vector]["data"].columns
                        and not calc_mode == "delta_ensembles"
                    ):
                        traces.append(
                            add_history_trace(
                                dfs[vector]["data"],
                                historical_vector_name,
                                line_shape,
                            )
                        )

                # Remove unwanted legends(only keep one for each ensemble)
                for trace in traces:
                    if trace.get("showlegend"):
                        if trace.get("legendgroup") in legends:
                            trace["showlegend"] = False
                        else:
                            legends.append(trace.get("legendgroup"))
                    fig.add_trace(trace, i + 1, 1)
                # Add observations
                if calc_mode != "delta_ensembles" and self.observations.get(vector):
                    for trace in add_observation_trace(self.observations.get(vector)):
                        fig.add_trace(trace, i + 1, 1)

            # Remove linked x-axis for histograms
            # Keep uirevision (e.g. zoom) for unchanged data.
            fig.update_xaxes(uirevision="locked")  # Time axis state kept
            for i, vector in enumerate(vectors, start=1):
                if "Histogram" in trace_options:
                    fig.update_xaxes(
                        row=i,
                        col=2,
                        matches=None,
                        showticklabels=True,
                        uirevision=vector,
                    )
                    fig.update_yaxes(row=i, col=2, uirevision=vector)
                fig.update_yaxes(row=i, col=1, uirevision=vector)

            fig = fig.to_dict()
            fig["layout"].update(
                barmode="overlay",
                bargap=0.01,
                bargroupgap=0.2,
            )
            fig["layout"] = self.theme.create_themed_layout(fig["layout"])
            return fig

        @app.callback(
            self.plugin_data_output,
            [self.plugin_data_requested],
            [
                State(self.uuid("vectors"), "selectedNodes"),
                State(self.uuid("ensemble"), "value"),
                State(self.uuid("mode"), "value"),
                State(self.uuid("base_ens"), "value"),
                State(self.uuid("delta_ens"), "value"),
                State(self.uuid("statistics"), "value"),
                State(self.uuid("cum_interval"), "value"),
                State(
                    self.uuid("vector_calculator_expressions"),
                    "data",
                ),
            ],
        )
        def _user_download_data(
            data_requested: Union[int, None],
            vectors: List[str],
            ensembles: List[str],
            calc_mode: str,
            base_ens: str,
            delta_ens: str,
            visualization: str,
            cum_interval: str,
            expressions: List[ExpressionInfo],
        ) -> Union[EncodedFile, str]:

            """Callback to download data based on selections"""
            if data_requested is None:
                raise PreventUpdate

            # Ensure selected ensembles is a list and prevent update if invalid calc_mode
            if calc_mode == "delta_ensembles":
                ensembles = [base_ens, delta_ens]
            elif calc_mode == "ensembles":
                if not isinstance(ensembles, list):
                    raise TypeError("ensembles should always be of type list")
            else:
                raise PreventUpdate

            if vectors is None:
                vectors = self.plot_options.get("vectors", [self.smry_cols[0]])

            # Calculate selected expressions:
            selected_expressions = get_selected_expressions(expressions, vectors)

            dfs = calculate_vector_dataframes(
                smry=self.smry,
                smry_meta=self.smry_meta,
                ensembles=ensembles,
                vectors=vectors,
                selected_expressions=selected_expressions,
                calc_mode=calc_mode,
                visualization=visualization,
                time_index=self.time_index,
                cum_interval=cum_interval,
            )
            for vector, df in dfs.items():
                if visualization in ["fanchart", "statistics"]:
                    df["stat"] = df["stat"].sort_values(
                        by=[("", "ENSEMBLE"), ("", "DATE")]
                    )
                    if vector.startswith(("AVG_", "INTVL_")):
                        df["stat"]["", "DATE"] = df["stat"]["", "DATE"].astype(str)
                        df["stat"]["", "DATE"] = df["stat"]["", "DATE"].apply(
                            date_to_interval_conversion,
                            vector=vector,
                            interval=cum_interval,
                            as_date=False,
                        )
                else:
                    df["data"] = df["data"].sort_values(by=["ENSEMBLE", "REAL", "DATE"])
                    # Reorder columns
                    df["data"] = df["data"][
                        ["ENSEMBLE", "REAL", "DATE"]
                        + [
                            col
                            for col in df["data"].columns
                            if col not in ["ENSEMBLE", "REAL", "DATE"]
                        ]
                    ]
                    if vector.startswith(("AVG_", "INTVL_")):
                        df["data"]["DATE"] = df["data"]["DATE"].astype(str)
                        df["data"]["DATE"] = df["data"]["DATE"].apply(
                            date_to_interval_conversion,
                            vector=vector,
                            interval=cum_interval,
                            as_date=False,
                        )

            # : is replaced with _ in filenames to stay within POSIX portable pathnames
            # (e.g. : is not valid in a Windows path)
            return WebvizPluginABC.plugin_data_compress(
                [
                    {
                        "filename": f"{vector.replace(':', '_')}.csv",
                        "content": df.get("stat", df["data"]).to_csv(index=False),
                    }
                    for vector, df in dfs.items()
                ]
            )

        @app.callback(
            [
                Output(self.uuid("show_ensembles"), "style"),
                Output(self.uuid("calc_delta"), "style"),
            ],
            [Input(self.uuid("mode"), "value")],
        )
        def _update_mode(mode: str) -> Tuple[dict, dict]:
            """Switch displayed ensemble selector for delta/no-delta"""
            if mode == "ensembles":
                style = {"display": "block"}, {"display": "none"}
            else:
                style = {"display": "none"}, {"display": "block"}
            return style

        @app.callback(
            Output(self.uuid("view_stat_options"), "style"),
            [Input(self.uuid("statistics"), "value")],
        )
        def _update_view_stat_options(visualization: str) -> dict:
            """Only show statistics picker if in statistics mode"""
            return (
                {"display": "block"}
                if visualization == "statistics"
                else {"display": "none"}
            )

        @app.callback(
            Output(self.uuid("date"), "data"),
            [Input(self.uuid("graph"), "clickData")],
            [State(self.uuid("date"), "data")],
        )
        def _update_date(clickdata: dict, date: str) -> str:
            """Store clicked date for use in other callback"""
            date = clickdata["points"][0]["x"] if clickdata else json.loads(date)
            return json.dumps(date)

        @app.callback(
            Output(self.uuid("cum_interval"), "options"),
            [
                Input(self.uuid("vectors"), "selectedNodes"),
            ],
            [State(self.uuid("cum_interval"), "options")],
        )
        def _activate_interval_radio_buttons(
            vectors: List[str],
            options: List[dict],
        ) -> List[dict]:
            """Switch activate/deactivate radio buttons for selectibg interval for
            calculations from cumulatives"""
            active = False
            if vectors:
                for vector in vectors:
                    if vector is not None and vector.startswith(("AVG_", "INTVL_")):
                        active = True
                        break
            if active:
                return [dict(option, **{"disabled": False}) for option in options]
            return [dict(option, **{"disabled": True}) for option in options]

        @app.callback(
            Output(self.uuid("vector_calculator"), "externalParseData"),
            Input(self.uuid("vector_calculator"), "externalParseExpression"),
        )
        def _parse_vector_calculator_expression(
            expression: ExpressionInfo,
        ) -> ExternalParseData:
            if expression is None:
                raise PreventUpdate

            return wsc.VectorCalculator.external_parse_data(expression)

        @app.callback(
            [
                Output(self.uuid("vector_calculator_expressions"), "data"),
                Output(self.uuid("vectors"), "data"),
                Output(self.uuid("vectors"), "selectedTags"),
                Output(self.uuid("vectors"), "customVectorDefinitions"),
            ],
            Input(self.uuid("modal_vector_calculator"), "is_open"),
            [
                State(self.uuid("vector_calculator_expressions_modal_open"), "data"),
                State(self.uuid("vector_calculator_expressions"), "data"),
                State(self.uuid("vectors"), "selectedNodes"),
                State(self.uuid("vectors"), "customVectorDefinitions"),
            ],
        )
        def _update_vector_calculator_expressions_actual(
            modal_open: bool,
            new_expressions: List[ExpressionInfo],
            existing_expressions: List[ExpressionInfo],
            selected_vectors: List[str],
            custom_vector_definitions: dict,
        ) -> list:
            if modal_open or (new_expressions == existing_expressions):
                raise PreventUpdate

            # Deep copy to prevent modifying self.vector_data
            vector_data = copy.deepcopy(self.vector_data)
            self._add_expressions_to_vector_data(vector_data, new_expressions)

            new_selected_vectors = self._get_valid_vector_selections(
                vector_data, selected_vectors, new_expressions, existing_expressions
            )

            # Prevent updates if selected vectors are unchanged
            if new_selected_vectors == selected_vectors:
                new_selected_vectors = dash.no_update

            new_custom_vector_definitions = get_vector_definitions_from_expressions(
                new_expressions
            )

            if new_custom_vector_definitions == custom_vector_definitions:
                new_custom_vector_definitions = dash.no_update

            return [
                new_expressions,
                vector_data,
                new_selected_vectors,
                new_custom_vector_definitions,
            ]

        @app.callback(
            Output(self.uuid("vector_calculator_expressions_modal_open"), "data"),
            Input(self.uuid("vector_calculator"), "expressions"),
        )
        def _update_vector_calculator_expressions(
            expressions: List[ExpressionInfo],
        ) -> list:
            valid_expressions: List[ExpressionInfo] = [
                elm for elm in expressions if elm["isValid"]
            ]

            return valid_expressions

        @app.callback(
            Output(self.uuid("modal_vector_calculator"), "is_open"),
            [
                Input(self.uuid("vector_calculator_open_btn"), "n_clicks"),
            ],
            [State(self.uuid("modal_vector_calculator"), "is_open")],
        )
        def _toggle_modal(n_open_clicks: int, is_open: bool) -> Optional[bool]:
            if n_open_clicks:
                return not is_open
            raise PreventUpdate

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        functions: List[Tuple[Callable, list]] = []
        if self.csvfile:
            functions.append((read_csv, [{"csv_file": self.csvfile}]))
        else:
            functions.extend(self.emodel.webvizstore)
        if self.obsfile:
            functions.append((get_path, [{"path": self.obsfile}]))
        if self.predefined_expressions_path:
            functions.append((get_path, [{"path": self.predefined_expressions_path}]))
        return functions


# pylint: disable = too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calculate_vector_dataframes(
    smry: pd.DataFrame,
    smry_meta: Union[pd.DataFrame, None],
    ensembles: List[str],
    vectors: List[str],
    selected_expressions: List[ExpressionInfo],
    calc_mode: str,
    visualization: str,
    time_index: str,
    cum_interval: str,
) -> dict:
    """Wraps cached function for individual vectors"""
    return {
        vector: calculate_vector_dataframe(
            smry=smry,
            smry_meta=smry_meta,
            ensembles=ensembles,
            vector=vector,
            selected_expressions=selected_expressions,
            calc_mode=calc_mode,
            visualization=visualization,
            time_index=time_index,
            cum_interval=cum_interval,
        )
        for vector in vectors
    }


# pylint: disable = too-many-arguments
@CACHE.memoize(timeout=CACHE.TIMEOUT)
def calculate_vector_dataframe(
    smry: pd.DataFrame,
    smry_meta: Union[pd.DataFrame, None],
    ensembles: List[str],
    vector: str,
    selected_expressions: List[ExpressionInfo],
    calc_mode: str,
    visualization: str,
    time_index: str,
    cum_interval: str,
) -> Dict[str, pd.DataFrame]:
    expression = get_expression_from_name(vector, selected_expressions)
    if expression:
        data = get_calculated_vector_df(expression, smry, ensembles)
    elif vector.startswith("AVG_"):
        total_vector = f"{vector[4:7] + vector[7:].replace('R', 'T', 1)}"
        data = filter_df(smry, ensembles, total_vector, smry_meta, calc_mode)
        data = calc_from_cumulatives(
            data=data,
            column_keys=total_vector,
            time_index=cum_interval,
            time_index_input=time_index,
            as_rate=True,
        )
        vector = rename_vec_from_cum(vector=vector[4:], as_rate=True)
    elif vector.startswith("INTVL_"):
        total_vector = vector.lstrip("INTVL_")
        data = filter_df(smry, ensembles, total_vector, smry_meta, calc_mode)
        data = calc_from_cumulatives(
            data=data,
            column_keys=total_vector,
            time_index=cum_interval,
            time_index_input=time_index,
            as_rate=False,
        )
    else:
        data = filter_df(smry, ensembles, vector, smry_meta, calc_mode)

    if calc_mode == "delta_ensembles":
        data = calculate_delta(data, ensembles[0], ensembles[1])

    output: Dict[str, pd.DataFrame] = {"data": data}
    if visualization in [
        "statistics",
        "statistics_hist",
        "fanchart",
        "fanchart_hist",
    ]:
        output["stat"] = calc_series_statistics(data, [vector])
    return output


def filter_df(
    df: pd.DataFrame,
    ensembles: List[str],
    vector: str,
    smry_meta: Union[pd.DataFrame, None],
    calc_mode: str,
) -> pd.DataFrame:
    """Filter dataframe for current vector. Include history
    vector if present"""
    columns = ["REAL", "ENSEMBLE", "DATE", vector]
    historical_vector_name = historical_vector(vector=vector, smry_meta=smry_meta)
    if (
        historical_vector_name
        and historical_vector_name in df.columns
        and not "delta" in calc_mode
    ):
        columns.append(historical_vector_name)
    fdf = df[columns].copy()
    if historical_vector_name in fdf.columns:
        fdf.loc[
            fdf["DATE"].astype("str") > str(datetime.datetime.now()),
            historical_vector_name,
        ] = np.nan
    return fdf.loc[fdf["ENSEMBLE"].isin(ensembles)]


def calculate_delta(df: pd.DataFrame, base_ens: str, delta_ens: str) -> pd.DataFrame:
    """Calculate delta between two ensembles"""
    base_df = (
        df.loc[df["ENSEMBLE"] == base_ens]
        .set_index(["DATE", "REAL"])
        .drop("ENSEMBLE", axis=1)
    )
    delta_df = (
        df.loc[df["ENSEMBLE"] == delta_ens]
        .set_index(["DATE", "REAL"])
        .drop("ENSEMBLE", axis=1)
    )
    dframe = base_df.sub(delta_df).reset_index()
    dframe["ENSEMBLE"] = f"({base_ens}) - ({delta_ens})"
    return dframe.dropna(axis=0, how="any")


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def add_histogram_traces(
    dframe: pd.DataFrame,
    vector: str,
    date: Union[str, None],
    colors: dict,
    interval: str,
) -> List[dict]:
    """Renders a histogram trace per ensemble for a given date"""
    dframe[("DATE")] = dframe[("DATE")].astype(str)
    date = date_to_interval_conversion(
        date=date, vector=vector, interval=interval, as_date=True
    )
    data = dframe.loc[dframe[("DATE")] == date]

    return [
        {
            "type": "histogram",
            "x": list(ens_df[vector]),
            "name": ensemble,
            "marker": {
                "color": colors.get(ensemble, colors[list(colors.keys())[0]]),
                "line": {"color": "black", "width": 1},
            },
            "showlegend": False,
            "legendgroup": ensemble,
        }
        for ensemble, ens_df in data.groupby(("ENSEMBLE"))
    ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def add_observation_trace(obs: dict) -> List[dict]:
    return [
        {
            "x": [value.get("date"), []],
            "y": [value.get("value"), []],
            "marker": {"color": "black"},
            "text": value.get("comment", None),
            "hoverinfo": "y+x+text",
            "showlegend": False,
            "error_y": {
                "type": "data",
                "array": [value.get("error"), []],
                "visible": True,
            },
        }
        for value in obs.get("observations", [])
    ]


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def add_realization_traces(
    dframe: pd.DataFrame, vector: str, colors: dict, line_shape: str, interval: str
) -> List[dict]:
    """Renders line trace for each realization, includes history line if present"""
    hovertemplate = render_hovertemplate(vector, interval)
    return [
        {
            "line": {"shape": line_shape},
            "x": list(real_df["DATE"]),
            "y": list(real_df[vector]),
            "hovertemplate": f"{hovertemplate}Realization: {real}, Ensemble: {ensemble}",
            "name": ensemble,
            "legendgroup": ensemble,
            "marker": {"color": colors.get(ensemble, colors[list(colors.keys())[0]])},
            "showlegend": real_no == 0,
        }
        for ens_no, (ensemble, ens_df) in enumerate(dframe.groupby("ENSEMBLE"))
        for real_no, (real, real_df) in enumerate(ens_df.groupby("REAL"))
    ]


def add_history_trace(dframe: pd.DataFrame, vector: str, line_shape: str) -> dict:
    """Renders the history line"""
    df = dframe.loc[
        (dframe["REAL"] == dframe["REAL"].unique()[0])
        & (dframe["ENSEMBLE"] == dframe["ENSEMBLE"].unique()[0])
    ]
    return {
        "line": {"shape": line_shape},
        "x": df["DATE"],
        "y": df[vector],
        "hovertext": "History",
        "hoverinfo": "y+x+text",
        "name": "History",
        "marker": {"color": "black"},
        "showlegend": True,
        "legendgroup": "History",
    }


@CACHE.memoize(timeout=CACHE.TIMEOUT)
def _get_fanchart_traces(
    stat_df: pd.DataFrame, vector: str, colors: dict, line_shape: str, interval: str
) -> list:
    """Add fanchart traces for selected vector"""
    traces = []
    for ensemble, ens_df in stat_df.groupby(("", "ENSEMBLE")):

        data = FanchartData(
            samples=ens_df[("", "DATE")].tolist(),
            low_high=LowHighData(
                low_data=ens_df[(vector, "low_p90")].values,
                low_name="P90",
                high_data=ens_df[(vector, "high_p10")].values,
                high_name="P10",
            ),
            minimum_maximum=MinMaxData(
                minimum=ens_df[(vector, "min")].values,
                maximum=ens_df[(vector, "max")].values,
            ),
            free_line=FreeLineData("Mean", ens_df[(vector, "mean")].values),
        )
        traces.extend(
            get_fanchart_traces(
                data=data,
                hex_color=colors.get(ensemble, colors[list(colors.keys())[0]]),
                legend_group=ensemble,
                line_shape=line_shape,
                hovertemplate=render_hovertemplate(vector, interval),
            )
        )
    return traces


def _add_statistics_traces(
    stat_df: pd.DataFrame,
    vector: str,
    colors: dict,
    line_shape: str,
    interval: str,
    stat_options: List[str],
) -> list:
    """Add fanchart traces for selected vector"""
    traces = []
    for ensemble, ens_df in stat_df.groupby(("", "ENSEMBLE")):
        traces.extend(
            add_statistics_traces(
                ens_stat_df=ens_df,
                vector=vector,
                color=colors.get(ensemble, colors[list(colors.keys())[0]]),
                legend_group=ensemble,
                line_shape=line_shape,
                hovertemplate=render_hovertemplate(vector=vector, interval=interval),
                stat_options=stat_options,
            )
        )
    return traces


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=False)


@webvizstore
def get_path(path: Path) -> Path:
    return Path(path)



import hashlib
import logging
import os
from pathlib import Path
from typing import List

from webviz_config.webviz_factory import WebvizFactory
from webviz_config.webviz_factory_registry import WEBVIZ_FACTORY_REGISTRY
from webviz_config.webviz_instance_info import WebvizRunMode

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._provider_impl_file import ProviderImplFile
from ._surface_discovery import (
    discover_observed_surface_files,
    discover_per_realization_surface_files,
)
from .ensemble_surface_provider import EnsembleSurfaceProvider

LOGGER = logging.getLogger(__name__)


class EnsembleSurfaceProviderFactory(WebvizFactory):
    def __init__(
        self,
        root_storage_folder: Path,
        allow_storage_writes: bool,
        avoid_copying_surfaces: bool,
    ) -> None:
        self._storage_dir = Path(root_storage_folder) / __name__
        self._allow_storage_writes = allow_storage_writes
        self._avoid_copying_surfaces = avoid_copying_surfaces

        LOGGER.info(
            f"EnsembleSurfaceProviderFactory init: storage_dir={self._storage_dir}"
        )

        if self._allow_storage_writes:
            os.makedirs(self._storage_dir, exist_ok=True)

    @staticmethod
    def instance() -> "EnsembleSurfaceProviderFactory":
        """Static method to access the singleton instance of the factory."""

        factory = WEBVIZ_FACTORY_REGISTRY.get_factory(EnsembleSurfaceProviderFactory)
        if not factory:
            app_instance_info = WEBVIZ_FACTORY_REGISTRY.app_instance_info
            storage_folder = app_instance_info.storage_folder
            allow_writes = app_instance_info.run_mode != WebvizRunMode.PORTABLE
            dont_copy_surfs = app_instance_info.run_mode == WebvizRunMode.NON_PORTABLE

            factory = EnsembleSurfaceProviderFactory(
                root_storage_folder=storage_folder,
                allow_storage_writes=allow_writes,
                avoid_copying_surfaces=dont_copy_surfs,
            )

            # Store the factory object in the global factory registry
            WEBVIZ_FACTORY_REGISTRY.set_factory(EnsembleSurfaceProviderFactory, factory)

        return factory

    def create_from_ensemble_surface_files(
        self,
        ens_path: str,
        rel_surface_folder: str = "share/results/maps",
        attribute_filter: List[str] = None,
    ) -> EnsembleSurfaceProvider:
        timer = PerfTimer()
        string_to_hash = (
            f"{ens_path}_{rel_surface_folder}"
            if attribute_filter is None
            else (
                f"{ens_path}_{rel_surface_folder}_"
                f"{'_'.join([str(attr) for attr in attribute_filter])}"
            )
        )
        storage_key = f"ens__{_make_hash_string(string_to_hash)}"
        provider = ProviderImplFile.from_backing_store(self._storage_dir, storage_key)
        if provider:
            LOGGER.info(
                f"Loaded surface provider from backing store in {timer.elapsed_s():.2f}s ("
                f"ens_path={ens_path})"
            )
            return provider

        # We can only import data from data source if storage writes are allowed
        if not self._allow_storage_writes:
            raise ValueError(f"Failed to load surface provider for {ens_path}")

        LOGGER.info(f"Importing/copying surface data for: {ens_path}")

        timer.lap_s()
        sim_surface_files = discover_per_realization_surface_files(
            ens_path, rel_surface_folder, attribute_filter
        )
        obs_surface_files = discover_observed_surface_files(ens_path, attribute_filter)
        et_discover_s = timer.lap_s()

        # As an optimization, avoid copying the surfaces into the backing store,
        # typically when  we're running in non-portable mode
        ProviderImplFile.write_backing_store(
            self._storage_dir,
            storage_key,
            sim_surfaces=sim_surface_files,
            obs_surfaces=obs_surface_files,
            avoid_copying_surfaces=self._avoid_copying_surfaces,
        )
        et_write_s = timer.lap_s()

        provider = ProviderImplFile.from_backing_store(self._storage_dir, storage_key)
        if not provider:
            raise ValueError(f"Failed to load/create surface provider for {ens_path}")

        LOGGER.info(
            f"Saved surface provider to backing store in {timer.elapsed_s():.2f}s ("
            f"discover={et_discover_s:.2f}s, write={et_write_s:.2f}s, ens_path={ens_path})"
        )

        return provider


def _make_hash_string(string_to_hash: str) -> str:
    # There is no security risk here and chances of collision should be very slim
    return hashlib.md5(string_to_hash.encode()).hexdigest()  # nosec
